# sage_setup: distribution = sagemath-graphs
r"""
Covering designs: coverings of `t`-element subsets of a `v`-set by `k`-sets

A `(v, k, t)` covering design `C` is an incidence structure consisting of a
set of points `P` of order `v`, and a set of blocks `B`, where each
block contains `k` points of `P`.  Every `t`-element subset of `P`
must be contained in at least one block.

If every `t`-set is contained in exactly one block of `C`, then we
have a block design.  Following the block design implementation, the
standard representation of a covering design uses `P = [0, 1, ..., v-1]`.

In addition to the parameters and incidence structure for a covering
design from this database, we include extra information:

* Best known lower bound on the size of a `(v, k, t)`-covering design
* Name of the person(s) who produced the design
* Method of construction used
* Date when the design was added to the database

REFERENCES:

.. [1] La Jolla Covering Repository,
  https://dmgordon.org/cover

.. [2] Daniel M. Gordon and Douglas R. Stinson, *Coverings*,
  Chapter 1 in: Charles J. Colbourn and Jeffrey H. Dinitz,
  *Handbook of Combinatorial Designs*, 2nd Edition, 2006.
  https://www.dmgordon.org/papers/hcd.pdf

AUTHORS:

- Daniel M. Gordon (2008-12-22): initial version

Classes and methods
-------------------
"""
# ****************************************************************************
#     Copyright (C) 2008 Daniel M. Gordon <dmgordo@gmail.com>
#
# Distributed under the terms of the GNU General Public License (GPL)
#
# The full text of the GPL is available at:
#
#                         https://www.gnu.org/licenses/
# ****************************************************************************

from urllib.request import urlopen
from ssl import create_default_context as default_context

from sage.structure.sage_object import SageObject
from sage.rings.rational import Rational
from sage.arith.misc import binomial
from sage.combinat.combination import Combinations
from sage.combinat.designs.incidence_structures import IncidenceStructure
from sage.cpython.string import bytes_to_str


def schonheim(v, k, t):
    r"""
    Return the Schonheim lower bound for the size of such a covering design.

    INPUT:

    - ``v`` -- integer; size of point set

    - ``k`` -- integer; cardinality of each block

    - ``t`` -- integer; cardinality of sets being covered

    OUTPUT:

    The Schonheim lower bound for such a covering design's size:
    `C(v, k, t) \leq \lceil(\frac{v}{k} \lceil \frac{v-1}{k-1} \cdots
    \lceil \frac{v-t+1}{k-t+1} \rceil \cdots \rceil \rceil`

    EXAMPLES::

        sage: from sage.combinat.designs.covering_design import schonheim
        sage: schonheim(10, 3, 2)
        17
        sage: schonheim(32, 16, 8)
        930
    """
    bound = 1
    for i in range(t - 1, -1, -1):
        bound = Rational((bound * (v - i), k - i)).ceil()
    return bound


def trivial_covering_design(v, k, t):
    r"""
    Construct a trivial covering design.

    INPUT:

    - ``v`` -- integer; size of point set

    - ``k`` -- integer; cardinality of each block

    - ``t`` -- integer; cardinality of sets being covered

    OUTPUT:

    A trivial `(v, k, t)` covering design

    EXAMPLES::

        sage: C = trivial_covering_design(8, 3, 1)
        sage: print(C)
        C(8, 3, 1) = 3
        Method: Trivial
        0   1   2
        0   6   7
        3   4   5
        sage: C = trivial_covering_design(5, 3, 2)
        sage: print(C)
        4 <= C(5, 3, 2) <= 10
        Method: Trivial
        0   1   2
        0   1   3
        0   1   4
        0   2   3
        0   2   4
        0   3   4
        1   2   3
        1   2   4
        1   3   4
        2   3   4

    .. NOTE::

        Cases are:

        * `t=0`: This could be empty, but it's a useful convention to have
          one block (which is empty if `k=0`).

        * `t=1` : This contains `\lceil v/k \rceil` blocks:
          `[0, ..., k-1], [k, ..., 2k-1], ...`.  The last block wraps around if
          `k` does not divide `v`.

        * anything else: Just use every `k`-subset of `[0, 1,..., v-1]`.
    """
    if t == 0:  # single block [0, ..., k-1]
        blk = list(range(k))
        return CoveringDesign(v, k, t, 1, range(v), [blk], 1, "Trivial")
    if t == 1:  # blocks [0, ..., k-1], [k, ..., 2k-1], ...
        size = Rational((v, k)).ceil()
        blocks = [list(range(i * k, (i + 1) * k))
                  for i in range(size - 1)]
        # last block: if k does not divide v, wrap around
        blk = list(range((size - 1) * k, v))
        for j in range(k - len(blk)):
            blk.append(j)
        blk.sort()
        blocks.append(blk)
        return CoveringDesign(v, k, t, size, range(v), blocks, size, "Trivial")
    # default case, all k-subsets
    return CoveringDesign(v, k, t, binomial(v, k), range(v),
                          Combinations(range(v), k), schonheim(v, k, t),
                          "Trivial")


class CoveringDesign(SageObject):
    """
    Covering design.

    INPUT:

    - ``v``, ``k``, ``t`` -- integer parameters of the covering design

    - ``size`` -- integer

    - ``points`` -- list of points (default: `[0, ..., v-1]`)

    - ``blocks``

    - ``low_bd`` -- integer; lower bound for such a design

    - ``method``, ``creator``, ``timestamp`` -- database information
    """

    def __init__(self, v=0, k=0, t=0, size=0, points=None, blocks=None,
                 low_bd=0, method='', creator='', timestamp=''):
        """
        EXAMPLES::

            sage: C = CoveringDesign(5, 4, 3, 4, range(5), [[0, 1, 2, 3],
            ....:     [0, 1, 2, 4], [0, 1, 3, 4], [0, 2, 3, 4]], 4,
            ....:     'Lexicographic Covering')
            sage: print(C)
            C(5, 4, 3) = 4
            Method: Lexicographic Covering
            0   1   2   3
            0   1   2   4
            0   1   3   4
            0   2   3   4
        """
        self.__v = v
        self.__k = k
        self.__t = t
        self.__size = size
        if low_bd > 0:
            self.__low_bd = low_bd
        else:
            self.__low_bd = schonheim(v, k, t)
        self.__method = method
        self.__creator = creator
        self.__timestamp = timestamp
        if points is None:
            points = []
        if blocks is None:
            blocks = []
        self.__incidence_structure = IncidenceStructure(points, blocks)

    def __repr__(self):
        """
        Return the representation of this covering design.

        This has the parameters but not the blocks.

        EXAMPLES::

            sage: C = CoveringDesign(7, 3, 2, 7, range(7), [[0, 1, 2],
            ....:     [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6],
            ....:     [2, 3, 6], [2, 4, 5]], 0, 'Projective Plane')
            sage: C
            (7, 3, 2)-covering design of size 7
            Lower bound: 7
            Method: Projective Plane
        """
        repr = ('(%d, %d, %d)-covering design of size %d\n'
                % (self.__v, self.__k, self.__t, self.__size))
        repr += 'Lower bound: %d\n' % (self.__low_bd)
        if self.__creator:
            repr += 'Created by: %s\n' % (self.__creator)
        if self.__method:
            repr += 'Method: %s\n' % (self.__method)
        if self.__timestamp:
            repr += 'Submitted on: %s\n' % (self.__timestamp)
        return repr[:-1]

    def __str__(self):
        """
        Return the string for this covering design.

        This has the parameters and the blocks in readable form.

        EXAMPLES::

            sage: C = CoveringDesign(7, 3, 2, 7, range(7), [[0, 1, 2],
            ....:     [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6],
            ....:     [2, 3, 6], [2, 4, 5]], 0, 'Projective Plane')
            sage: print(C)
            C(7, 3, 2) = 7
            Method: Projective Plane
            0   1   2
            0   3   4
            0   5   6
            1   3   5
            1   4   6
            2   3   6
            2   4   5
        """
        if self.__size == self.__low_bd:  # check if covering is optimal
            repr = ('C(%d, %d, %d) = %d\n' %
                    (self.__v, self.__k, self.__t, self.__size))
        else:
            repr = ('%d <= C(%d, %d, %d) <= %d\n' % (self.__low_bd,
                    self.__v, self.__k, self.__t, self.__size))
        if self.__creator:
            repr += 'Created by: %s\n' % (self.__creator)
        if self.__method:
            repr += 'Method: %s\n' % (self.__method)
        if self.__timestamp:
            repr += 'Submitted on: %s\n' % (self.__timestamp)
        return repr + '\n'.join('  '.join(str(k) for k in block) for block in
                                self.__incidence_structure.blocks())

    def is_covering(self):
        """
        Check all `t`-sets are in fact covered by the blocks of ``self``.

        .. NOTE::

            This is very slow and wasteful of memory.

        EXAMPLES::

            sage: C = CoveringDesign(7, 3, 2, 7, range(7), [[0, 1, 2],
            ....:     [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6],
            ....:     [2, 3, 6], [2, 4, 5]], 0, 'Projective Plane')
            sage: C.is_covering()
            True
            sage: C = CoveringDesign(7, 3, 2, 7, range(7), [[0, 1, 2],
            ....:     [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6], [2, 3, 6],
            ....:     [2, 4, 6]], 0, 'not a covering')   # last block altered
            sage: C.is_covering()
            False
        """
        v = self.__v
        k = self.__k
        t = self.__t
        Svt = Combinations(range(v), t)
        Skt = Combinations(range(k), t)
        tset = {}  # tables of t-sets: False = uncovered, True = covered
        for i in Svt:
            tset[tuple(i)] = False
        # mark all t-sets covered by each block
        for a in self.__incidence_structure.blocks():
            for z in Skt:
                y = (a[x] for x in z)
                tset[tuple(y)] = True
        for i in Svt:
            if not tset[tuple(i)]:  # uncovered
                return False
        return True                 # everything was covered

    def v(self):
        """
        Return `v`, the number of points in the covering design.

        EXAMPLES::

            sage: from sage.combinat.designs.covering_design import CoveringDesign
            sage: C = CoveringDesign(7, 3, 2, 7, range(7), [[0, 1, 2],
            ....:     [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6],
            ....:     [2, 3, 6], [2, 4, 5]], 0, 'Projective Plane')
            sage: C.v()
            7
        """
        return self.__v

    def k(self):
        """
        Return `k`, the size of blocks of the covering design.

        EXAMPLES::

            sage: from sage.combinat.designs.covering_design import CoveringDesign
            sage: C = CoveringDesign(7, 3, 2, 7, range(7), [[0, 1, 2],
            ....:     [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6],
            ....:     [2, 3, 6], [2, 4, 5]], 0, 'Projective Plane')
            sage: C.k()
            3
        """
        return self.__k

    def t(self):
        """
        Return `t`, the size of sets which must be covered by the
        blocks of the covering design

        EXAMPLES::

            sage: from sage.combinat.designs.covering_design import CoveringDesign
            sage: C = CoveringDesign(7, 3, 2, 7, range(7), [[0, 1, 2],
            ....:     [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6],
            ....:     [2, 3, 6], [2, 4, 5]], 0, 'Projective Plane')
            sage: C.t()
            2
        """
        return self.__t

    def size(self):
        """
        Return the number of blocks in the covering design.

        EXAMPLES::

            sage: from sage.combinat.designs.covering_design import CoveringDesign
            sage: C = CoveringDesign(7, 3, 2, 7, range(7), [[0, 1, 2],
            ....:     [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6],
            ....:     [2, 3, 6], [2, 4, 5]], 0, 'Projective Plane')
            sage: C.size()
            7
        """
        return self.__size

    def low_bd(self):
        """
        Return a lower bound for the number of blocks a covering
        design with these parameters could have.

        Typically this is the Schonheim bound, but for some parameters
        better bounds have been shown.

        EXAMPLES::

            sage: from sage.combinat.designs.covering_design import CoveringDesign
            sage: C = CoveringDesign(7, 3, 2, 7, range(7), [[0, 1, 2],
            ....:     [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6],
            ....:     [2, 3, 6], [2, 4, 5]], 0, 'Projective Plane')
            sage: C.low_bd()
            7
        """
        return self.__low_bd

    def method(self):
        """
        Return the method used to create the covering design.

        This field is optional, and is used in a database to give information
        about how coverings were constructed.

        EXAMPLES::

            sage: from sage.combinat.designs.covering_design import CoveringDesign
            sage: C = CoveringDesign(7, 3, 2, 7, range(7), [[0, 1, 2],
            ....:     [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6],
            ....:     [2, 3, 6], [2, 4, 5]], 0, 'Projective Plane')
            sage: C.method()
            'Projective Plane'
        """
        return self.__method

    def creator(self):
        """
        Return the creator of the covering design.

        This field is optional, and is used in a database to give
        attribution for the covering design It can refer to the person
        who submitted it, or who originally gave a construction

        EXAMPLES::

            sage: from sage.combinat.designs.covering_design import CoveringDesign
            sage: C = CoveringDesign(7, 3, 2, 7, range(7), [[0, 1, 2],
            ....:     [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6], [2, 3, 6],
            ....:     [2, 4, 5]],0, 'Projective Plane', 'Gino Fano')
            sage: C.creator()
            'Gino Fano'
        """
        return self.__creator

    def timestamp(self):
        """
        Return the time that the covering was submitted to the database.

        EXAMPLES::

            sage: from sage.combinat.designs.covering_design import CoveringDesign
            sage: C = CoveringDesign(7, 3, 2, 7, range(7), [[0, 1, 2],
            ....:     [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6],
            ....:     [2, 3, 6], [2, 4, 5]],0, 'Projective Plane',
            ....:     'Gino Fano', '1892-01-01 00:00:00')
            sage: C.timestamp()  # No exact date known; in Fano's 1892 article
            '1892-01-01 00:00:00'
        """
        return self.__timestamp

    def incidence_structure(self):
        """
        Return the incidence structure of this design, without extra parameters.

        EXAMPLES::

            sage: from sage.combinat.designs.covering_design import CoveringDesign
            sage: C = CoveringDesign(7, 3, 2, 7, range(7), [[0, 1, 2],
            ....:     [0, 3, 4], [0, 5, 6], [1, 3, 5], [1, 4, 6],
            ....:     [2, 3, 6], [2, 4, 5]], 0, 'Projective Plane')
            sage: D = C.incidence_structure()
            sage: D.ground_set()
            [0, 1, 2, 3, 4, 5, 6]
            sage: D.blocks()
            [[0, 1, 2], [0, 3, 4], [0, 5, 6], [1, 3, 5],
            [1, 4, 6], [2, 3, 6], [2, 4, 5]]
        """
        return self.__incidence_structure


def best_known_covering_design_www(v, k, t, verbose=False):
    r"""
    Return the best known `(v, k, t)` covering design from an online database.

    This uses the La Jolla Covering Repository, a database
    available at `<https://ljcr.dmgordon.org/cover.html>`_

    INPUT:

    - ``v`` -- integer; the size of the point set for the design

    - ``k`` -- integer; the number of points per block

    - ``t`` -- integer; the size of sets covered by the blocks

    - ``verbose`` -- boolean (default: ``False``); print verbose message

    OUTPUT:

    A :class:`CoveringDesign` object representing the ``(v, k, t)``-covering
    design with smallest number of blocks available in the database.

    EXAMPLES::

        sage: from sage.combinat.designs.covering_design import (  # optional - internet
        ....:     best_known_covering_design_www)
        sage: C = best_known_covering_design_www(7, 3, 2)  # optional - internet
        sage: print(C)                                     # optional - internet
        C(7, 3, 2) = 7
        Method: lex covering
        Submitted on: 1996-12-01 00:00:00
        0  1  2
        0  3  4
        0  5  6
        1  3  5
        1  4  6
        2  3  6
        2  4  5

    A :exc:`ValueError` is raised if the ``(v, k, t)`` parameters are not
    found in the database.
    """
    from sage.misc.sage_eval import sage_eval

    v = int(v)
    k = int(k)
    t = int(t)
    param = "?v=%s&k=%s&t=%s" % (v, k, t)
    url = "https://ljcr.dmgordon.org/get_cover.php" + param
    if verbose:
        print("Looking up the bounds at %s" % url)

    f = urlopen(url, context=default_context())
    try:
        s = bytes_to_str(f.read())
    finally:
        f.close()

    if 'covering not in database' in s:  # not found
        str = "no (%d, %d, %d) covering design in database\n" % (v, k, t)
        raise ValueError(str)
    return sage_eval(s)



# sage_setup: distribution = sagemath-graphs
#*****************************************************************************
#       Copyright (C) 2008-2009 Robert L. Miller <rlmillster@gmail.com>
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  http://www.gnu.org/licenses/
#*****************************************************************************

from sage.graphs.base.c_graph cimport CGraph, CGraphBackend
from sage.data_structures.binary_matrix cimport binary_matrix_t

cdef class DenseGraph(CGraph):
    cdef bint _directed
    cdef binary_matrix_t edges
    cdef inline int _add_arc_unsafe(self, int, int) except -1
    cdef inline int _del_arc_unsafe(self, int u, int v) except -1

cdef int copy_dense_graph(DenseGraph dest, DenseGraph src) except -1

cdef class DenseGraphBackend(CGraphBackend):
    cdef DenseGraph _cg
    cdef inline CGraph cg(self):
        return <CGraph> self._cg



# sage_setup: distribution = sagemath-graphs

cdef class TreeIterator:
    cdef int vertices
    cdef int first_time
    cdef int p
    cdef int q
    cdef int h1
    cdef int h2
    cdef int c
    cdef int r
    cdef int *l
    cdef int *current_level_sequence
    cdef int generate_first_level_sequence(self) noexcept
    cdef int generate_next_level_sequence(self) noexcept



# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.graphs
r"""
Examples of simplicial sets.

These are accessible via ``simplicial_sets.Sphere(3)``,
``simplicial_sets.Torus()``, etc.  Type ``simplicial_sets.[TAB]`` to
see a complete list.

AUTHORS:

- John H. Palmieri (2016-07)

- Miguel Marco (2022-12)
"""
# ****************************************************************************
#  Copyright (C) 2016 John H. Palmieri <palmieri at math.washington.edu>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#                  https://www.gnu.org/licenses/
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty
#    of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
#
#  See the GNU General Public License for more details; the full text
#  is available at:
#
#                  https://www.gnu.org/licenses/
#
# ****************************************************************************

import re
from pathlib import Path

from sage.env import SAGE_ENV
from sage.misc.cachefunc import cached_method, cached_function
from sage.misc.latex import latex
from sage.rings.infinity import Infinity
from sage.rings.integer import Integer
from sage.structure.parent import Parent

from .delta_complex import delta_complexes
from .simplicial_set import AbstractSimplex, \
    SimplicialSet_arbitrary, SimplicialSet_finite

import sage.topology.simplicial_complex_catalog as simplicial_complexes

from sage.misc.lazy_import import lazy_import
lazy_import('sage.categories.simplicial_sets', 'SimplicialSets')

kenzo_path = Path(SAGE_ENV['SAGE_EXTCODE']) / 'kenzo'


# ######################################################################
# The nerve of a finite monoid, used in sage.categories.finite_monoid.

class Nerve(SimplicialSet_arbitrary):
    def __init__(self, monoid):
        """
        The nerve of a multiplicative monoid.

        INPUT:

        - ``monoid`` -- a multiplicative monoid

        See
        :meth:`sage.categories.finite_monoids.FiniteMonoids.ParentMethods.nerve`
        for full documentation.

        EXAMPLES::

            sage: M = FiniteMonoids().example()
            sage: M
            An example of a finite multiplicative monoid: the integers modulo 12
            sage: X = M.nerve()
            sage: list(M)
            [0, 1, 2, 3, 4, 5, 6, 7, 8, 9, 10, 11]
            sage: X.n_cells(0)
            [1]
            sage: X.n_cells(1)
            [0, 10, 11, 2, 3, 4, 5, 6, 7, 8, 9]
        """
        category = SimplicialSets().Pointed()
        Parent.__init__(self, category=category)
        self.rename("Nerve of {}".format(str(monoid)))
        self.rename_latex("B{}".format(latex(monoid)))

        e = AbstractSimplex(0, name=str(monoid.one()),
                            latex_name=latex(monoid.one()))
        self._basepoint = e
        vertex = SimplicialSet_finite({e: None}, base_point=e)
        # self._n_skeleton: cache the highest dimensional skeleton
        # calculated so far for this simplicial set, along with its
        # dimension.
        self._n_skeleton = (0, vertex)
        self._monoid = monoid
        # self._simplex_data: a tuple whose elements are pairs (simplex, list
        # of monoid elements). Omit the base point.
        self._simplex_data = ()

    def __eq__(self, other) -> bool:
        """
        Return ``True`` if ``self`` and ``other`` are equal.

        This checks that the underlying monoids and the underlying
        base points are the same. Because the base points will be
        different each time the nerve is constructed, different
        instances will not be equal.

        EXAMPLES::

            sage: # needs sage.groups
            sage: C3 = groups.misc.MultiplicativeAbelian([3])
            sage: C3.nerve() == C3.nerve()
            False
            sage: BC3 = C3.nerve()
            sage: BC3 == BC3
            True
        """
        return (isinstance(other, Nerve)
                and self._monoid == other._monoid
                and self.base_point() == other.base_point())

    def __ne__(self, other) -> bool:
        """
        Return the negation of ``__eq__``.

        EXAMPLES::

            sage: # needs sage.groups
            sage: C3 = groups.misc.MultiplicativeAbelian([3])
            sage: G3 = groups.permutation.Cyclic(3)
            sage: C3.nerve() != G3.nerve()
            True
            sage: C3.nerve() != C3.nerve()
            True
        """
        return not self == other

    @cached_method
    def __hash__(self):
        """
        The hash is formed from the monoid and the base point.

        EXAMPLES::

            sage: G3 = groups.permutation.Cyclic(3)                                     # needs sage.groups
            sage: hash(G3.nerve()) # random                                             # needs sage.groups
            17

        Different instances yield different base points, hence different hashes::

            sage: # needs sage.groups
            sage: X = G3.nerve()
            sage: Y = G3.nerve()
            sage: X.base_point() != Y.base_point()
            True
            sage: hash(X) != hash(Y)
            True
        """
        return hash(self._monoid) ^ hash(self.base_point())

    def n_skeleton(self, n):
        """
        Return the `n`-skeleton of this simplicial set.

        That is, the simplicial set generated by all nondegenerate
        simplices of dimension at most `n`.

        INPUT:

        - ``n`` -- the dimension

        EXAMPLES::

            sage: # needs sage.groups
            sage: K4 = groups.misc.MultiplicativeAbelian([2,2])
            sage: BK4 = simplicial_sets.ClassifyingSpace(K4)
            sage: BK4.n_skeleton(3)
            Simplicial set with 40 non-degenerate simplices
            sage: BK4.n_cells(1) == BK4.n_skeleton(3).n_cells(1)
            True
            sage: BK4.n_cells(3) == BK4.n_skeleton(1).n_cells(3)
            False
        """
        from .simplicial_set_constructions import SubSimplicialSet
        monoid = self._monoid
        one = monoid.one()
        # Build up chains of elements inductively, from dimension d-1
        # to dimension d. We start with the cached
        # self._n_skeleton. If only the 0-skeleton has been
        # constructed, we construct the 1-cells by hand.
        start, skel = self._n_skeleton
        if start == n:
            return skel
        elif start > n:
            return skel.n_skeleton(n)

        # There is a single vertex. Name it after the identity
        # element of the monoid.
        e = skel.n_cells(0)[0]
        # Build the dictionary simplices, to be used for
        # constructing the simplicial set.
        simplices = skel.face_data()

        # face_dict: dictionary of simplices: keys are
        # composites of monoid elements (as tuples), values are
        # the corresponding simplices.
        face_dict = dict(self._simplex_data)

        if start == 0:
            for g in monoid:
                if g != one:
                    x = AbstractSimplex(1, name=str(g), latex_name=latex(g))
                    simplices[x] = (e, e)
                    face_dict[(g,)] = x
            start = 1

        for d in range(start + 1, n + 1):
            for g in monoid:
                if g == one:
                    continue
                new_faces = {}
                for t in face_dict.keys():
                    if len(t) != d - 1:
                        continue
                    # chain: chain of group elements to multiply,
                    # as a tuple.
                    chain = t + (g,)
                    # bdries: the face maps applied to chain, in a
                    # format suitable for passing to the DeltaComplex
                    # constructor.
                    x = AbstractSimplex(d,
                                        name=' * '.join(str(_) for _ in chain),
                                        latex_name=' * '.join(latex(_) for _ in chain))
                    new_faces[chain] = x

                    # Compute faces of x.
                    faces = [face_dict[chain[1:]]]
                    for i in range(d - 1):
                        product = chain[i] * chain[i + 1]
                        if product == one:
                            # Degenerate.
                            if d == 2:
                                face = e.apply_degeneracies(i)
                            else:
                                face = (face_dict[chain[:i]
                                                  + chain[i+2:]].apply_degeneracies(i))
                        else:
                            # Non-degenerate.
                            face = (face_dict[chain[:i]
                                              + (product,) + chain[i+2:]])
                        faces.append(face)
                    faces.append(face_dict[chain[:-1]])
                    simplices[x] = faces
                face_dict.update(new_faces)

        K = SubSimplicialSet(simplices, self)
        self._n_skeleton = (n, K)
        self._simplex_data = face_dict.items()
        return K


########################################################################
# Catalog of examples. These are accessed via simplicial_set_catalog.py.

def Sphere(n):
    r"""
    Return the `n`-sphere as a simplicial set.

    It is constructed with two non-degenerate simplices: a vertex
    `v_0` (which is the base point) and an `n`-simplex `\sigma_n`.

    INPUT:

    - ``n`` -- integer

    EXAMPLES::

        sage: S0 = simplicial_sets.Sphere(0)
        sage: S0
        S^0
        sage: S0.nondegenerate_simplices()
        [v_0, w_0]
        sage: S0.is_pointed()
        True
        sage: simplicial_sets.Sphere(4)
        S^4
        sage: latex(simplicial_sets.Sphere(4))
        S^{4}
        sage: simplicial_sets.Sphere(4).nondegenerate_simplices()
        [v_0, sigma_4]
    """
    v_0 = AbstractSimplex(0, name='v_0')
    if n == 0:
        w_0 = AbstractSimplex(0, name='w_0')
        return SimplicialSet_finite({v_0: None, w_0: None}, base_point=v_0,
                                    name='S^0')
    degens = range(n - 2, -1, -1)
    degen_v = v_0.apply_degeneracies(*degens)
    sigma = AbstractSimplex(n, name='sigma_{}'.format(n),
                            latex_name='\\sigma_{}'.format(n))
    return SimplicialSet_finite({sigma: [degen_v] * (n + 1)}, base_point=v_0,
                                name='S^{}'.format(n),
                                latex_name='S^{{{}}}'.format(n))


def ClassifyingSpace(group):
    r"""
    Return the classifying space of ``group``, as a simplicial set.

    INPUT:

    - ``group`` -- a finite group or finite monoid

    See
    :meth:`sage.categories.finite_monoids.FiniteMonoids.ParentMethods.nerve`
    for more details and more examples.

    EXAMPLES::

        sage: # needs sage.groups
        sage: C2 = groups.misc.MultiplicativeAbelian([2])
        sage: BC2 = simplicial_sets.ClassifyingSpace(C2)
        sage: H = BC2.homology(range(9), base_ring=GF(2))                               # needs sage.modules
        sage: [H[i].dimension() for i in range(9)]                                      # needs sage.modules
        [0, 1, 1, 1, 1, 1, 1, 1, 1]

        sage: Klein4 = groups.misc.MultiplicativeAbelian([2, 2])                        # needs sage.groups
        sage: BK = simplicial_sets.ClassifyingSpace(Klein4); BK                         # needs sage.groups
        Classifying space of Multiplicative Abelian group isomorphic to C2 x C2
        sage: BK.homology(range(5), base_ring=GF(2))    # long time (1 second)          # needs sage.groups sage.modules
        {0: Vector space of dimension 0 over Finite Field of size 2,
         1: Vector space of dimension 2 over Finite Field of size 2,
         2: Vector space of dimension 3 over Finite Field of size 2,
         3: Vector space of dimension 4 over Finite Field of size 2,
         4: Vector space of dimension 5 over Finite Field of size 2}
    """
    X = group.nerve()
    X.rename('Classifying space of {}'.format(group))
    return X


def RealProjectiveSpace(n):
    r"""
    Return real `n`-dimensional projective space, as a simplicial set.

    This is constructed as the `n`-skeleton of the nerve of the group
    of order 2, and therefore has a single non-degenerate simplex in
    each dimension up to `n`.

    EXAMPLES::

        sage: # needs sage.groups
        sage: simplicial_sets.RealProjectiveSpace(7)
        RP^7
        sage: RP5 = simplicial_sets.RealProjectiveSpace(5)
        sage: RP5.homology()
        {0: 0, 1: C2, 2: 0, 3: C2, 4: 0, 5: Z}
        sage: RP5
        RP^5
        sage: latex(RP5)
        RP^{5}

        sage: BC2 = simplicial_sets.RealProjectiveSpace(Infinity)                       # needs sage.groups
        sage: latex(BC2)                                                                # needs sage.groups
        RP^{\infty}
    """
    if n == Infinity:
        from sage.groups.abelian_gps.abelian_group import AbelianGroup

        X = AbelianGroup([2]).nerve()
        X.rename('RP^oo')
        X.rename_latex('RP^{\\infty}')
    else:
        X = RealProjectiveSpace(Infinity).n_skeleton(n)
        X.rename('RP^{}'.format(n))
        X.rename_latex('RP^{{{}}}'.format(n))
    return X


def KleinBottle():
    r"""
    Return the Klein bottle as a simplicial set.

    This converts the `\Delta`-complex version to a simplicial set. It
    has one 0-simplex, three 1-simplices, and two 2-simplices.

    EXAMPLES::

        sage: K = simplicial_sets.KleinBottle()
        sage: K.f_vector()
        [1, 3, 2]
        sage: K.homology(reduced=False)                                                 # needs sage.modules
        {0: Z, 1: Z x C2, 2: 0}
        sage: K
        Klein bottle
    """
    temp = SimplicialSet_finite(delta_complexes.KleinBottle())
    pt = temp.n_cells(0)[0]
    return SimplicialSet_finite(temp.face_data(), base_point=pt,
                                name='Klein bottle')


def Torus():
    r"""
    Return the torus as a simplicial set.

    This computes the product of the circle with itself, where the
    circle is represented using a single 0-simplex and a single
    1-simplex. Thus it has one 0-simplex, three 1-simplices, and two
    2-simplices.

    EXAMPLES::

        sage: T = simplicial_sets.Torus()
        sage: T.f_vector()
        [1, 3, 2]
        sage: T.homology(reduced=False)                                                 # needs sage.modules
        {0: Z, 1: Z x Z, 2: Z}
    """
    S1 = Sphere(1)
    T = S1.product(S1)
    T.rename('Torus')
    return T


def Simplex(n):
    r"""
    Return the `n`-simplex as a simplicial set.

    EXAMPLES::

        sage: K = simplicial_sets.Simplex(2)
        sage: K
        2-simplex
        sage: latex(K)
        \Delta^{2}
        sage: K.n_cells(0)
        [(0,), (1,), (2,)]
        sage: K.n_cells(1)
        [(0, 1), (0, 2), (1, 2)]
        sage: K.n_cells(2)
        [(0, 1, 2)]
    """
    return SimplicialSet_finite(simplicial_complexes.Simplex(n),
                                name='{}-simplex'.format(n),
                                latex_name='\\Delta^{{{}}}'.format(n))


@cached_function
def Empty():
    """
    Return the empty simplicial set.

    This should return the same simplicial set each time it is called.

    EXAMPLES::

        sage: from sage.topology.simplicial_set_examples import Empty
        sage: E = Empty()
        sage: E
        Empty simplicial set
        sage: E.nondegenerate_simplices()
        []
        sage: E is Empty()
        True
    """
    return SimplicialSet_finite({}, name='Empty simplicial set')


@cached_function
def Point():
    """
    Return a single point called "*" as a simplicial set.

    This should return the same simplicial set each time it is called.

    EXAMPLES::

        sage: P = simplicial_sets.Point()
        sage: P.is_pointed()
        True
        sage: P.nondegenerate_simplices()
        [*]

        sage: Q = simplicial_sets.Point()
        sage: P is Q
        True
        sage: P == Q
        True
    """
    star = AbstractSimplex(0, name='*')
    return SimplicialSet_finite({star: None}, base_point=star,
                                name='Point',
                                latex_name='*')


def Horn(n, k):
    r"""
    Return the horn `\Lambda^n_k`.

    This is the subsimplicial set of the `n`-simplex obtained by
    removing its `k`-th face.

    EXAMPLES::

        sage: L = simplicial_sets.Horn(3, 0)
        sage: L
        (3, 0)-Horn
        sage: L.n_cells(3)
        []
        sage: L.n_cells(2)
        [(0, 1, 2), (0, 1, 3), (0, 2, 3)]

        sage: L20 = simplicial_sets.Horn(2, 0)
        sage: latex(L20)
        \Lambda^{2}_{0}
        sage: L20.inclusion_map()
        Simplicial set morphism:
          From: (2, 0)-Horn
          To:   2-simplex
          Defn: [(0,), (1,), (2,), (0, 1), (0, 2)] --> [(0,), (1,), (2,), (0, 1), (0, 2)]
    """
    K = Simplex(n)
    sigma = K.n_cells(n)[0]
    L = K.subsimplicial_set(K.faces(sigma)[:k] + K.faces(sigma)[k+1:])
    L.rename('({}, {})-Horn'.format(n, k))
    L.rename_latex('\\Lambda^{{{}}}_{{{}}}'.format(n, k))
    return L


def ComplexProjectiveSpace(n):
    r"""
    Return complex `n`-dimensional projective space, as a simplicial set.

    This is only defined when `n` is at most 4. It is constructed
    using the simplicial set decomposition provided by Kenzo, as
    described by Sergeraert [Ser2010]_

    EXAMPLES::

        sage: simplicial_sets.ComplexProjectiveSpace(2).homology(reduced=False)         # needs sage.modules
        {0: Z, 1: 0, 2: Z, 3: 0, 4: Z}
        sage: CP3 = simplicial_sets.ComplexProjectiveSpace(3); CP3                      # needs pyparsing
        CP^3
        sage: latex(CP3)                                                                # needs pyparsing
        CP^{3}
        sage: CP3.f_vector()                                                            # needs pyparsing
        [1, 0, 3, 10, 25, 30, 15]

        sage: # long time, needs pyparsing sage.modules
        sage: K = CP3.suspension()              # long time (1 second)
        sage: R = K.cohomology_ring(GF(2))
        sage: R.gens()
        (h^{0,0}, h^{3,0}, h^{5,0}, h^{7,0})
        sage: x = R.gens()[1]
        sage: x.Sq(2)
        h^{5,0}

        sage: simplicial_sets.ComplexProjectiveSpace(4).f_vector()                      # needs pyparsing
        [1, 0, 4, 22, 97, 255, 390, 315, 105]

        sage: simplicial_sets.ComplexProjectiveSpace(5)
        Traceback (most recent call last):
        ...
        ValueError: complex projective spaces are only available in dimensions between 0 and 4
    """
    if n < 0 or n > 4:
        raise ValueError('complex projective spaces are only available in dimensions between 0 and 4')
    if n == 0:
        return Point()
    if n == 1:
        return Sphere(2)
    if n == 2:
        # v: Kenzo name <<GBar>>
        v = AbstractSimplex(0, name='v')
        # f_2_i: Kenzo name <<GBar<- (i)><- NIL>>> for i=1,2
        f2_1 = AbstractSimplex(2, name='rho_0')
        f2_2 = AbstractSimplex(2, name='rho_1')
        # f3_110: Kenzo name <<GBar<- (1 1)><0 NIL><- NIL>>>
        # f3_011: Kenzo name <<GBar<0 (1)><- (1)><- NIL>>>
        # f3_111: Kenzo name <<GBar<1 (1)><- (1)><- NIL>>>
        f3_110 = AbstractSimplex(3, name='sigma_0', latex_name='\\sigma_0')
        f3_011 = AbstractSimplex(3, name='sigma_1', latex_name='\\sigma_1')
        f3_111 = AbstractSimplex(3, name='sigma_2', latex_name='\\sigma_2')
        # f4_101101: Kenzo name <<GBar<1-0 (1)><1-0 NIL><- (1)><- NIL>>>
        # f4_201110: Kenzo name <<GBar<2-0 (1)><1 (1)><0 NIL><- NIL>>>
        # f4_211010: Kenzo name <<GBar<2-1 (1)><0 (1)><0 NIL><- NIL>>>
        f4_101101 = AbstractSimplex(4, name='tau_0', latex_name='\\tau_0')
        f4_201110 = AbstractSimplex(4, name='tau_1', latex_name='\\tau_1')
        f4_211010 = AbstractSimplex(4, name='tau_2', latex_name='\\tau_2')
        K = SimplicialSet_finite({f2_1: (v.apply_degeneracies(0),
                                         v.apply_degeneracies(0),
                                         v.apply_degeneracies(0)),
                                  f2_2: (v.apply_degeneracies(0),
                                         v.apply_degeneracies(0),
                                         v.apply_degeneracies(0)),
                                  f3_110: (f2_1, f2_2, f2_1, v.apply_degeneracies(1, 0)),
                                  f3_011: (f2_1, f2_1, f2_1, f2_1),
                                  f3_111: (v.apply_degeneracies(1, 0), f2_1, f2_2, f2_1),
                                  f4_101101: (f2_1.apply_degeneracies(0),
                                              f2_1.apply_degeneracies(0),
                                              f3_011,
                                              f2_1.apply_degeneracies(2),
                                              f2_1.apply_degeneracies(2)),
                                  f4_201110: (f2_1.apply_degeneracies(1),
                                              f3_111,
                                              f3_011,
                                              f3_110,
                                              f2_1.apply_degeneracies(1)),
                                  f4_211010: (f2_1.apply_degeneracies(2),
                                              f3_111,
                                              f2_1.apply_degeneracies(1),
                                              f3_110,
                                              f2_1.apply_degeneracies(0))},
                                 base_point=v, name='CP^2',
                                 latex_name='CP^{2}')
        return K
    if n == 3:
        file = kenzo_path / 'CP3.txt'
        data = simplicial_data_from_kenzo_output(file)
        v = [sigma for sigma in data if sigma.dimension() == 0][0]
        return SimplicialSet_finite(data, base_point=v, name='CP^3',
                                    latex_name='CP^{3}')
    if n == 4:
        file = kenzo_path / 'CP4.txt'
        data = simplicial_data_from_kenzo_output(file)
        v = [sigma for sigma in data if sigma.dimension() == 0][0]
        return SimplicialSet_finite(data, base_point=v, name='CP^4',
                                    latex_name='CP^{4}')


def simplicial_data_from_kenzo_output(filename) -> dict:
    """
    Return data to construct a simplicial set, given Kenzo output.

    INPUT:

    - ``filename`` -- name of file containing the output from Kenzo's
      :func:`show-structure` function

    OUTPUT: data to construct a simplicial set from the Kenzo output

    Several files with Kenzo output are in the directory
    :file:`SAGE_EXTCODE/kenzo/`.

    EXAMPLES::

        sage: from sage.topology.simplicial_set_examples import simplicial_data_from_kenzo_output
        sage: from sage.topology.simplicial_set import SimplicialSet
        sage: from pathlib import Path
        sage: sphere = Path(SAGE_ENV['SAGE_EXTCODE']) / 'kenzo' /'S4.txt'
        sage: S4 = SimplicialSet(simplicial_data_from_kenzo_output(sphere))             # needs pyparsing
        sage: S4.homology(reduced=False)                                                # needs pyparsing
        {0: Z, 1: 0, 2: 0, 3: 0, 4: Z}
    """
    from pyparsing import OneOrMore, nestedExpr

    with open(filename) as f:
        data = f.read()
    dim = 0
    start = 0
    # simplex_data: data for constructing the simplicial set.
    simplex_data = {}
    # simplex_names: simplices indexed by their names
    simplex_names = {}
    dim_idx = data.find('Dimension = {}:'.format(dim), start)
    while dim_idx != -1:
        start = dim_idx + len('Dimension = {}:'.format(dim))
        new_dim_idx = data.find('Dimension = {}:'.format(dim + 1), start)
        if new_dim_idx == -1:
            end = len(data)
        else:
            end = new_dim_idx
        if dim == 0:
            simplex_string = data[data.find('Vertices :') + len('Vertices :'):end]
            vertices = OneOrMore(nestedExpr()).parseString(simplex_string).asList()[0]
            for v in vertices:
                vertex = AbstractSimplex(0, name=v)
                simplex_data[vertex] = None
                simplex_names[v] = vertex
        else:
            simplex_string = data[start:end].strip()

            for s in [_.strip() for _ in simplex_string.split('Simplex : ')]:
                if s:
                    name, face_str = (_.strip() for _ in s.split('Faces : '))
                    face_str = face_str.strip('()')
                    face_str = face_str.split('<AbSm ')
                    faces = []
                    for f in face_str[1:]:
                        # f has the form 'DEGENS NAME>', possibly with a trailing space.
                        # DEGENS is a hyphen-separated list, like
                        # '3-2-1-0' or '0' or '-'.
                        m = re.match('[-[0-9]+', f)
                        degen_str = m.group(0)
                        if degen_str.find('-') != -1:
                            if degen_str == '-':
                                degens = []
                            else:
                                degens = [Integer(_)
                                          for _ in degen_str.split('-')]
                        else:
                            degens = [Integer(degen_str)]

                        face_name = f[m.end(0):].strip()[:-1]
                        nondegen = simplex_names[face_name]
                        faces.append(nondegen.apply_degeneracies(*degens))

                    simplex = AbstractSimplex(dim, name=name)
                    simplex_data[simplex] = faces
                    simplex_names[name] = simplex
        dim += 1
        dim_idx = new_dim_idx
    return simplex_data


def HopfMap():
    r"""
    Return a simplicial model of the Hopf map `S^3 \to S^2`.

    This is taken from Exemple II.1.19 in the thesis of Clemens Berger
    [Ber1991]_.

    The Hopf map is a fibration `S^3 \to S^2`. If it is viewed as
    attaching a 4-cell to the 2-sphere, the resulting adjunction space
    is 2-dimensional complex projective space. The resulting model is
    a bit larger than the one obtained from
    ``simplicial_sets.ComplexProjectiveSpace(2)``.

    EXAMPLES::

        sage: g = simplicial_sets.HopfMap()
        sage: g.domain()
        Simplicial set with 20 non-degenerate simplices
        sage: g.codomain()
        S^2

    Using the Hopf map to attach a cell::

        sage: X = g.mapping_cone()
        sage: CP2 = simplicial_sets.ComplexProjectiveSpace(2)
        sage: X.homology() == CP2.homology()                                            # needs sage.modules
        True

        sage: X.f_vector()
        [1, 0, 5, 9, 6]
        sage: CP2.f_vector()
        [1, 0, 2, 3, 3]
    """
    # The 2-sphere and its simplices.
    S2 = Sphere(2)
    sigma = S2.n_cells(2)[0]
    s0_sigma = sigma.apply_degeneracies(0)
    s1_sigma = sigma.apply_degeneracies(1)
    s2_sigma = sigma.apply_degeneracies(2)
    # The 3-sphere and its simplices.
    w_0 = AbstractSimplex(0, name='w')
    w_1 = w_0.apply_degeneracies(0)
    w_2 = w_0.apply_degeneracies(0, 0)
    beta_11 = AbstractSimplex(1, name='beta_11', latex_name='\\beta_{11}')
    beta_22 = AbstractSimplex(1, name='beta_22', latex_name='\\beta_{22}')
    beta_23 = AbstractSimplex(1, name='beta_23', latex_name='\\beta_{23}')
    beta_44 = AbstractSimplex(1, name='beta_44', latex_name='\\beta_{44}')
    beta_1 = AbstractSimplex(2, name='beta_1', latex_name='\\beta_1')
    beta_2 = AbstractSimplex(2, name='beta_2', latex_name='\\beta_2')
    beta_3 = AbstractSimplex(2, name='beta_3', latex_name='\\beta_3')
    beta_4 = AbstractSimplex(2, name='beta_4', latex_name='\\beta_4')
    alpha_12 = AbstractSimplex(2, name='alpha_12', latex_name='\\alpha_{12}')
    alpha_23 = AbstractSimplex(2, name='alpha_23', latex_name='\\alpha_{23}')
    alpha_34 = AbstractSimplex(2, name='alpha_34', latex_name='\\alpha_{34}')
    alpha_45 = AbstractSimplex(2, name='alpha_45', latex_name='\\alpha_{45}')
    alpha_56 = AbstractSimplex(2, name='alpha_56', latex_name='\\alpha_{56}')
    alpha_1 = AbstractSimplex(3, name='alpha_1', latex_name='\\alpha_1')
    alpha_2 = AbstractSimplex(3, name='alpha_2', latex_name='\\alpha_2')
    alpha_3 = AbstractSimplex(3, name='alpha_3', latex_name='\\alpha_3')
    alpha_4 = AbstractSimplex(3, name='alpha_4', latex_name='\\alpha_4')
    alpha_5 = AbstractSimplex(3, name='alpha_5', latex_name='\\alpha_5')
    alpha_6 = AbstractSimplex(3, name='alpha_6', latex_name='\\alpha_6')
    S3 = SimplicialSet_finite({beta_11: (w_0, w_0), beta_22: (w_0, w_0),
                               beta_23: (w_0, w_0), beta_44: (w_0, w_0),
                               beta_1: (w_1, beta_11, w_1),
                               beta_2: (w_1, beta_22, beta_23),
                               beta_3: (w_1, beta_23, w_1),
                               beta_4: (w_1, beta_44, w_1),
                               alpha_12: (beta_11, beta_23, w_1),
                               alpha_23: (beta_11, beta_22, w_1),
                               alpha_34: (beta_11, beta_22, beta_44),
                               alpha_45: (w_1, beta_23, beta_44),
                               alpha_56: (w_1, beta_23, w_1),
                               alpha_1: (beta_1, beta_3, alpha_12, w_2),
                               alpha_2: (beta_11.apply_degeneracies(1), beta_2,
                                         alpha_23, alpha_12),
                               alpha_3: (beta_11.apply_degeneracies(0), alpha_34,
                                         alpha_23, beta_4),
                               alpha_4: (beta_1, beta_2, alpha_34, alpha_45),
                               alpha_5: (w_2, alpha_45, alpha_56, beta_4),
                               alpha_6: (w_2, beta_3, alpha_56, w_2)},
                              base_point=w_0)
    return S3.Hom(S2)({alpha_1: s0_sigma, alpha_2: s1_sigma,
                       alpha_3: s2_sigma, alpha_4: s0_sigma,
                       alpha_5: s2_sigma, alpha_6: s1_sigma})


def PresentationComplex(G):
    r"""
    Return a simplicial set constructed from a group presentation.
    The result is a subdivision of the presentation complex.

    The presentation complex has a single vertex and it has one edge for
    each generator. Then triangles (and eventually new edges
    to glue them) are added to realize the relations.

    INPUT:

    - ``G`` -- a finitely presented group

    EXAMPLES::

        sage: # needs sage.groups
        sage: G = SymmetricGroup(2).as_finitely_presented_group(); G
        Finitely presented group < a | a^2 >
        sage: S = simplicial_sets.PresentationComplex(G); S
        Simplicial set with 5 non-degenerate simplices
        sage: S.face_data()
        {Delta^0: None,
         a: (Delta^0, Delta^0),
         a^-1: (Delta^0, Delta^0),
         Ta: (a, s_0 Delta^0, a^-1),
         a^2: (a, s_0 Delta^0, a)}
        sage: S.fundamental_group()
        Finitely presented group < e0 | e0^2 >
    """
    O = AbstractSimplex(0)
    SO = O.apply_degeneracies(0)
    edges = {g: AbstractSimplex(1, name=str(g)) for g in G.gens()}
    inverseedges = {g.inverse(): AbstractSimplex(1, name=str(g.inverse())) for g in G.gens()}
    all_edges = {}
    all_edges.update(edges)
    all_edges.update(inverseedges)
    triangles = {g: AbstractSimplex(2, name='T' + str(g)) for g in G.gens()}
    face_maps = {g: [O, O] for g in all_edges.values()}
    face_maps.update({triangles[t]: [all_edges[t], SO, all_edges[t.inverse()]] for t in triangles})
    for r in G.relations():
        if len(r.Tietze()) == 1:
            pass
        elif len(r.Tietze()) == 2:
            a = all_edges[G([r.Tietze()[0]])]
            b = all_edges[G([r.Tietze()[1]])]
            T = AbstractSimplex(2, name=str(r))
            face_maps[T] = [a, SO, b]
        else:
            words = [all_edges[G([a])] for a in r.Tietze()]
            words[-1] = all_edges[G([-r.Tietze()[-1]])]
            while len(words) > 3:
                auxedge = AbstractSimplex(1)
                face_maps[auxedge] = [O, O]
                auxtring = AbstractSimplex(2)
                face_maps[auxtring] = [words[1], auxedge, words[0]]
                words = [auxedge] + words[2:]
            auxtring = AbstractSimplex(2)
            face_maps[auxtring] = [words[1], words[2], words[0]]
    return SimplicialSet_finite(face_maps, base_point=O)



# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.modules
r"""
Families of graphs derived from classical geometries over finite fields

These include graphs of polar spaces, affine polar graphs, graphs
related to Hermitean unitals, graphs on nonisotropic points, etc.

The methods defined here appear in :mod:`sage.graphs.graph_generators`.
"""

# ****************************************************************************
#           Copyright (C) 2015 Sagemath project
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 2 of the License, or
# (at your option) any later version.
#                  https://www.gnu.org/licenses/
# ****************************************************************************

from sage.graphs.graph import Graph
from sage.arith.misc import is_prime_power
from sage.rings.finite_rings.finite_field_constructor import FiniteField


def SymplecticPolarGraph(d, q, algorithm=None):
    r"""
    Return the Symplectic Polar Graph `Sp(d,q)`.

    The Symplectic Polar Graph `Sp(d,q)` is built from a projective space of
    dimension `d-1` over a field `F_q`, and a symplectic form `f`. Two vertices
    `u,v` are made adjacent if `f(u,v)=0`.

    See the page `on symplectic graphs on Andries Brouwer's website
    <https://www.win.tue.nl/~aeb/graphs/Sp.html>`_.

    INPUT:

    - ``d``, ``q`` -- integers; note that only even values of `d` are accepted
      by the function

    - ``algorithm`` -- string (default: ``None``); if set to ``'gap'``, then the
      computation is carried via GAP library interface, computing totally
      singular subspaces, which is faster for `q>3`.  Otherwise it is done
      directly.

    EXAMPLES:

    Computation of the spectrum of `Sp(6,2)`::

        sage: g = graphs.SymplecticPolarGraph(6, 2)
        sage: g.is_strongly_regular(parameters=True)
        (63, 30, 13, 15)
        sage: set(g.spectrum()) == {-5, 3, 30}                                          # needs sage.rings.number_field
        True

    The parameters of `Sp(4,q)` are the same as of `O(5,q)`, but they are
    not isomorphic if `q` is odd::

        sage: G = graphs.SymplecticPolarGraph(4, 3)
        sage: G.is_strongly_regular(parameters=True)
        (40, 12, 2, 4)

        sage: # needs sage.libs.gap
        sage: O = graphs.OrthogonalPolarGraph(5, 3)
        sage: O.is_strongly_regular(parameters=True)
        (40, 12, 2, 4)
        sage: O.is_isomorphic(G)
        False
        sage: S = graphs.SymplecticPolarGraph(6, 4, algorithm='gap')    # not tested (long time)
        sage: S.is_strongly_regular(parameters=True)                    # not tested (long time)
        (1365, 340, 83, 85)

    TESTS::

        sage: graphs.SymplecticPolarGraph(4,4,algorithm='gap').is_strongly_regular(parameters=True)                     # needs sage.libs.gap
        (85, 20, 3, 5)
        sage: graphs.SymplecticPolarGraph(4,4).is_strongly_regular(parameters=True)     # needs sage.libs.pari
        (85, 20, 3, 5)
        sage: graphs.SymplecticPolarGraph(4,4,algorithm='blah')
        Traceback (most recent call last):
        ...
        ValueError: unknown algorithm!
    """
    if d < 1 or d % 2:
        raise ValueError("d must be even and greater than 2")

    if algorithm == "gap":     # faster for larger (q>3)  fields
        from sage.libs.gap.libgap import libgap
        G = _polar_graph(d, q, libgap.SymplecticGroup(d, q))

    elif algorithm is None:    # faster for small (q<4) fields
        from sage.modules.free_module import VectorSpace
        from sage.schemes.projective.projective_space import ProjectiveSpace
        from sage.matrix.constructor import identity_matrix, block_matrix, zero_matrix

        F = FiniteField(q, "x")
        M = block_matrix(F, 2, 2,
                         [zero_matrix(F, d/2),
                          identity_matrix(F, d/2),
                          -identity_matrix(F, d/2),
                          zero_matrix(F, d/2)])

        V = VectorSpace(F, d)
        PV = list(ProjectiveSpace(d - 1, F))
        G = Graph([[tuple(_) for _ in PV], lambda x, y: V(x)*(M*V(y)) == 0], loops=False)

    else:
        raise ValueError("unknown algorithm!")

    G.name("Symplectic Polar Graph Sp({},{})".format(d, q))
    G.relabel()
    return G


def AffineOrthogonalPolarGraph(d, q, sign='+'):
    r"""
    Return the affine polar graph `VO^+(d,q),VO^-(d,q)` or `VO(d,q)`.

    Affine Polar graphs are built from a `d`-dimensional vector space over
    `F_q`, and a quadratic form which is hyperbolic, elliptic or parabolic
    according to the value of ``sign``.

    Note that `VO^+(d,q),VO^-(d,q)` are strongly regular graphs, while `VO(d,q)`
    is not.

    For more information on Affine Polar graphs, see `Affine Polar Graphs page
    of Andries Brouwer's website <https://www.win.tue.nl/~aeb/graphs/VO.html>`_.

    INPUT:

    - ``d`` -- integer; ``d`` must be even if ``sign is not None``, and odd
      otherwise

    - ``q`` -- integer; a power of a prime number, as `F_q` must exist

    - ``sign`` -- string (default: ``'+'``); must be equal to ``'+'``, ``'-'``,
      or ``None`` to compute (respectively) `VO^+(d,q),VO^-(d,q)` or
      `VO(d,q)`

    .. NOTE::

        The graph `VO^\epsilon(d,q)` is the graph induced by the
        non-neighbors of a vertex in an :meth:`Orthogonal Polar Graph
        <OrthogonalPolarGraph>` `O^\epsilon(d+2,q)`.

    EXAMPLES:

    The :meth:`Brouwer-Haemers graph <BrouwerHaemersGraph>` is isomorphic to
    `VO^-(4,3)`::

        sage: g = graphs.AffineOrthogonalPolarGraph(4,3,"-")                            # needs sage.libs.gap
        sage: g.is_isomorphic(graphs.BrouwerHaemersGraph())                             # needs sage.libs.gap
        True

    Some examples from `Brouwer's table or strongly regular graphs
    <https://www.win.tue.nl/~aeb/graphs/srg/srgtab.html>`_::

        sage: # needs sage.libs.gap
        sage: g = graphs.AffineOrthogonalPolarGraph(6,2,"-"); g
        Affine Polar Graph VO^-(6,2): Graph on 64 vertices
        sage: g.is_strongly_regular(parameters=True)
        (64, 27, 10, 12)
        sage: g = graphs.AffineOrthogonalPolarGraph(6,2,"+"); g
        Affine Polar Graph VO^+(6,2): Graph on 64 vertices
        sage: g.is_strongly_regular(parameters=True)
        (64, 35, 18, 20)

    When ``sign is None``::

        sage: # needs sage.libs.gap
        sage: g = graphs.AffineOrthogonalPolarGraph(5,2,None); g
        Affine Polar Graph VO^-(5,2): Graph on 32 vertices
        sage: g.is_strongly_regular(parameters=True)
        False
        sage: g.is_regular()
        True
        sage: g.is_vertex_transitive()
        True
    """
    if sign in ["+", "-"]:
        s = 1 if sign == "+" else -1
        if d % 2:
            raise ValueError("d must be even when sign is not None")
    else:
        if not d % 2:
            raise ValueError("d must be odd when sign is None")
        s = 0

    from sage.modules.free_module import VectorSpace
    from sage.matrix.constructor import Matrix
    from sage.libs.gap.libgap import libgap
    from itertools import combinations

    M = Matrix(libgap.InvariantQuadraticForm(libgap.GeneralOrthogonalGroup(s, d, q))['matrix'])
    F = libgap.GF(q).sage()
    V = list(VectorSpace(F, d))

    G = Graph()
    G.add_vertices([tuple(_) for _ in V])
    for x, y in combinations(V, 2):
        if not (x - y)*M*(x - y):
            G.add_edge(tuple(x), tuple(y))

    G.name("Affine Polar Graph VO^" + str('+' if s == 1 else '-') + "(" + str(d) + "," + str(q) + ")")
    G.relabel()
    return G


def _orthogonal_polar_graph(m, q, sign='+', point_type=[0]):
    r"""
    A helper function to build ``OrthogonalPolarGraph`` and ``NO2,3,5`` graphs.

    See the `page of
    Andries Brouwer's website <https://www.win.tue.nl/~aeb/graphs/srghub.html>`_.

    INPUT:

    - ``m``, ``q`` -- integers; `q` must be a prime power

    - ``sign`` -- string (default: ``'+'``); must be ``'+'`` or ``'-'`` if `m`
      is even, ``'+'`` (default) otherwise

    - ``point_type`` -- list of elements from `F_q`

    EXAMPLES:

    Petersen graph::

        sage: from sage.graphs.generators.classical_geometries import _orthogonal_polar_graph
        sage: g = _orthogonal_polar_graph(3,5,point_type=[2,3])                         # needs sage.libs.gap
        sage: g.is_strongly_regular(parameters=True)                                    # needs sage.libs.gap
        (10, 3, 0, 1)

    A locally Petersen graph (a.k.a. Doro graph, a.k.a. Hall graph)::

        sage: g = _orthogonal_polar_graph(4,5,'-',point_type=[2,3])                     # needs sage.libs.gap
        sage: g.is_distance_regular(parameters=True)                                    # needs sage.libs.gap
        ([10, 6, 4, None], [None, 1, 2, 5])

    Various big and slow to build graphs:

    `NO^+(7,3)`::

        sage: g = _orthogonal_polar_graph(7,3,point_type=[1])    # not tested (long time)
        sage: g.is_strongly_regular(parameters=True)             # not tested (long time)
        (378, 117, 36, 36)

    `NO^-(7,3)`::

        sage: g = _orthogonal_polar_graph(7,3,point_type=[-1])   # not tested (long time)
        sage: g.is_strongly_regular(parameters=True)             # not tested (long time)
        (351, 126, 45, 45)

    `NO^+(6,3)`::

        sage: g = _orthogonal_polar_graph(6,3,point_type=[1])                           # needs sage.libs.gap
        sage: g.is_strongly_regular(parameters=True)                                    # needs sage.libs.gap
        (117, 36, 15, 9)

    `NO^-(6,3)`::

        sage: g = _orthogonal_polar_graph(6,3,'-',point_type=[1])                       # needs sage.libs.gap
        sage: g.is_strongly_regular(parameters=True)                                    # needs sage.libs.gap
        (126, 45, 12, 18)

    `NO^{-,\perp}(5,5)`::

        sage: g = _orthogonal_polar_graph(5,5,point_type=[2,3])         # long time, needs sage.libs.gap
        sage: g.is_strongly_regular(parameters=True)                    # long time, needs sage.libs.gap
        (300, 65, 10, 15)

    `NO^{+,\perp}(5,5)`::

        sage: g = _orthogonal_polar_graph(5,5,point_type=[1,-1]) # not tested (long time)
        sage: g.is_strongly_regular(parameters=True)             # not tested (long time)
        (325, 60, 15, 10)

    TESTS::

        sage: # needs sage.libs.gap
        sage: g = _orthogonal_polar_graph(5,3,point_type=[-1])
        sage: g.is_strongly_regular(parameters=True)
        (45, 12, 3, 3)
        sage: g = _orthogonal_polar_graph(5,3,point_type=[1])
        sage: g.is_strongly_regular(parameters=True)
        (36, 15, 6, 6)
    """
    from sage.schemes.projective.projective_space import ProjectiveSpace
    from sage.modules.free_module_element import free_module_element as vector
    from sage.matrix.constructor import Matrix
    from sage.libs.gap.libgap import libgap

    if not m % 2:
        if sign != "+" and sign != "-":
            raise ValueError("sign must be equal to either '-' or '+' when "
                             "m is even")
    else:
        if sign != "" and sign != "+":
            raise ValueError("sign must be equal to either '' or '+' when "
                             "m is odd")
        sign = ""

    e = {'+': 1,
         '-': -1,
         '': 0}[sign]

    M = Matrix(libgap.InvariantQuadraticForm(libgap.GeneralOrthogonalGroup(e, m, q))['matrix'])
    Fq = libgap.GF(q).sage()
    PG = [vector(s) for s in ProjectiveSpace(m - 1, Fq)]

    for v in PG:
        v.set_immutable()

    def F(x):
        return x*M*x

    if not q % 2:
        def P(x, y):
            return F(x - y)
    else:
        def P(x, y):
            return x*M*y + y*M*x

    V = [x for x in PG if F(x) in point_type]

    G = Graph([V, lambda x, y: P(x, y) == 0], loops=False)

    G.relabel()
    return G


def OrthogonalPolarGraph(m, q, sign='+'):
    r"""
    Return the Orthogonal Polar Graph `O^{\epsilon}(m,q)`.

    For more information on Orthogonal Polar graphs, see the `page of Andries
    Brouwer's website <https://www.win.tue.nl/~aeb/graphs/srghub.html>`_.

    INPUT:

    - ``m``, ``q`` -- integers; `q` must be a prime power

    - ``sign`` -- string (default: ``'+'``); must be ``'+'`` or ``'-'`` if `m`
      is even, ``'+'`` (default) otherwise

    EXAMPLES::

        sage: # needs sage.libs.gap
        sage: G = graphs.OrthogonalPolarGraph(6,3,"+"); G
        Orthogonal Polar Graph O^+(6, 3): Graph on 130 vertices
        sage: G.is_strongly_regular(parameters=True)
        (130, 48, 20, 16)
        sage: G = graphs.OrthogonalPolarGraph(6,3,"-"); G
        Orthogonal Polar Graph O^-(6, 3): Graph on 112 vertices
        sage: G.is_strongly_regular(parameters=True)
        (112, 30, 2, 10)
        sage: G = graphs.OrthogonalPolarGraph(5,3); G
        Orthogonal Polar Graph O(5, 3): Graph on 40 vertices
        sage: G.is_strongly_regular(parameters=True)
        (40, 12, 2, 4)
        sage: G = graphs.OrthogonalPolarGraph(8,2,"+"); G
        Orthogonal Polar Graph O^+(8, 2): Graph on 135 vertices
        sage: G.is_strongly_regular(parameters=True)
        (135, 70, 37, 35)
        sage: G = graphs.OrthogonalPolarGraph(8,2,"-"); G
        Orthogonal Polar Graph O^-(8, 2): Graph on 119 vertices
        sage: G.is_strongly_regular(parameters=True)
        (119, 54, 21, 27)

    TESTS::

        sage: G = graphs.OrthogonalPolarGraph(4,3,"")                                   # needs sage.libs.gap
        Traceback (most recent call last):
        ...
        ValueError: sign must be equal to either '-' or '+' when m is even
        sage: G = graphs.OrthogonalPolarGraph(5,3,"-")                                  # needs sage.libs.gap
        Traceback (most recent call last):
        ...
        ValueError: sign must be equal to either '' or '+' when m is odd
    """
    G = _orthogonal_polar_graph(m, q, sign=sign)
    if m % 2:
        sign = ""
    G.name("Orthogonal Polar Graph O" + ("^" + sign if sign else "") + str((m, q)))
    return G


def NonisotropicOrthogonalPolarGraph(m, q, sign='+', perp=None):
    r"""
    Return the Graph `NO^{\epsilon,\perp}_{m}(q)`.

    Let the vectorspace of dimension `m` over `F_q` be endowed with a
    nondegenerate quadratic form `F`, of type ``sign`` for `m` even.

    * `m` even: assume further that `q=2` or `3`. Returns the graph of the
      points (in the underlying projective space) `x` satisfying `F(x)=1`, with
      adjacency given by orthogonality w.r.t. `F`. Parameter ``perp`` is
      ignored.

    * `m` odd: if ``perp`` is not ``None``, then we assume that `q=5` and return
      the graph of the points `x` satisfying `F(x)=\pm 1` if ``sign="+"``,
      respectively `F(x) \in \{2,3\}` if ``sign="-"``, with adjacency given by
      orthogonality w.r.t. `F` (cf. Sect 7.D of [BL1984]_). Otherwise return
      the graph of nongenerate hyperplanes of type ``sign``, adjacent whenever
      the intersection is degenerate (cf. Sect. 7.C of [BL1984]_).
      Note that for `q=2` one will get a complete graph.

    For more information, see Sect. 9.9 of [BH2012]_ and [BL1984]_. Note that
    the `page of Andries Brouwer's website
    <https://www.win.tue.nl/~aeb/graphs/srghub.html>`_ uses different notation.

    INPUT:

    - ``m`` -- integer;  half the dimension of the underlying vectorspace

    - ``q`` -- a power of a prime number, the size of the underlying field

    - ``sign`` -- string (default: ``'+'``); must be either ``'+'`` or ``'-'``

    EXAMPLES:

    `NO^-(4,2)` is isomorphic to Petersen graph::

        sage: g = graphs.NonisotropicOrthogonalPolarGraph(4,2,'-'); g                   # needs sage.libs.gap
        NO^-(4, 2): Graph on 10 vertices
        sage: g.is_strongly_regular(parameters=True)                                    # needs sage.libs.gap
        (10, 3, 0, 1)

    `NO^-(6,2)` and `NO^+(6,2)`::

        sage: # needs sage.libs.gap
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(6,2,'-')
        sage: g.is_strongly_regular(parameters=True)
        (36, 15, 6, 6)
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(6,2,'+'); g
        NO^+(6, 2): Graph on 28 vertices
        sage: g.is_strongly_regular(parameters=True)
        (28, 15, 6, 10)

    `NO^+(8,2)`::

        sage: g = graphs.NonisotropicOrthogonalPolarGraph(8,2,'+')                      # needs sage.libs.gap
        sage: g.is_strongly_regular(parameters=True)                                    # needs sage.libs.gap
        (120, 63, 30, 36)

    Wilbrink's graphs for `q=5`::

        sage: # needs sage.libs.gap
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(5,5,perp=1)
        sage: g.is_strongly_regular(parameters=True)    # long time
        (325, 60, 15, 10)
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(5,5,'-',perp=1)
        sage: g.is_strongly_regular(parameters=True)    # long time
        (300, 65, 10, 15)

    Wilbrink's graphs::

        sage: # needs sage.libs.gap
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(5,4,'+')
        sage: g.is_strongly_regular(parameters=True)
        (136, 75, 42, 40)
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(5,4,'-')
        sage: g.is_strongly_regular(parameters=True)
        (120, 51, 18, 24)
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(7,4,'+'); g        # not tested (long time)
        NO^+(7, 4): Graph on 2080 vertices
        sage: g.is_strongly_regular(parameters=True)                         # not tested (long time)
        (2080, 1071, 558, 544)

    TESTS::

        sage: # needs sage.libs.gap
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(4,2); g
        NO^+(4, 2): Graph on 6 vertices
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(4,3,'-')
        sage: g.is_strongly_regular(parameters=True)
        (15, 6, 1, 3)
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(3,5,'-',perp=1); g
        NO^-,perp(3, 5): Graph on 10 vertices
        sage: g.is_strongly_regular(parameters=True)
        (10, 3, 0, 1)

        sage: # long time, needs sage.libs.gap
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(6,3,'+')
        sage: g.is_strongly_regular(parameters=True)
        (117, 36, 15, 9)
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(6,3,'-'); g
        NO^-(6, 3): Graph on 126 vertices
        sage: g.is_strongly_regular(parameters=True)
        (126, 45, 12, 18)
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(5,5,'-')
        sage: g.is_strongly_regular(parameters=True)
        (300, 104, 28, 40)
        sage: g = graphs.NonisotropicOrthogonalPolarGraph(5,5,'+')
        sage: g.is_strongly_regular(parameters=True)
        (325, 144, 68, 60)

        sage: g = graphs.NonisotropicOrthogonalPolarGraph(6,4,'+')
        Traceback (most recent call last):
        ...
        ValueError: for m even q must be 2 or 3
    """
    p, k = is_prime_power(q, get_data=True)
    if not k:
        raise ValueError('q must be a prime power')
    dec = ''
    if not m % 2:
        if q in [2, 3]:
            G = _orthogonal_polar_graph(m, q, sign=sign, point_type=[1])
        else:
            raise ValueError("for m even q must be 2 or 3")
    elif perp is not None:
        if q == 5:
            pt = [-1, 1] if sign == '+' else [2, 3] if sign == '-' else []
            G = _orthogonal_polar_graph(m, q, point_type=pt)
            dec = ",perp"
        else:
            raise ValueError("for perp not None q must be 5")
    else:
        if sign not in ['+', '-']:
            raise ValueError("sign must be '+' or '-'")
        from sage.libs.gap.libgap import libgap
        g0 = libgap.GeneralOrthogonalGroup(m, q)
        g = libgap.Group(libgap.List(g0.GeneratorsOfGroup(), libgap.TransposedMat))
        F = libgap.GF(q)  # F_q
        W = libgap.FullRowSpace(F, m)  # F_q^m
        e = 1 if sign == '+' else -1
        n = (m - 1)/2
        # we build (q^n(q^n+e)/2, (q^n-e)(q^(n-1)+e), 2(q^(2n-2)-1)+eq^(n-1)(q-1),
        #                                          2q^(n-1)(q^(n-1)+e))-srg
        # **use** v and k to select appropriate orbit and orbital
        nvert = (q**n)*(q**n + e)/2  # v
        deg = (q**n - e)*(q**(n - 1) + e)   # k
        S = [libgap.Elements(libgap.Basis(x))[0]
             for x in libgap.Elements(libgap.Subspaces(W, 1))]
        (V,) = (x for x in libgap.Orbits(g, S, libgap.OnLines)
                if len(x) == nvert)
        gp = libgap.Action(g, V, libgap.OnLines)  # make a permutation group
        h = libgap.Stabilizer(gp, 1)
        (Vh,) = (x for x in libgap.Orbits(h, libgap.Orbit(gp, 1))
                 if len(x) == deg)
        Vh = Vh[0]
        L = libgap.Orbit(gp, [1, Vh], libgap.OnSets)
        G = Graph()
        G.add_edges(L)
    G.name("NO^" + sign + dec + str((m, q)))
    return G


def _polar_graph(m, q, g, intersection_size=None):
    r"""
    The helper function to build graphs `(D)U(m,q)` and `(D)Sp(m,q)`.

    Building a graph on an orbit of a group `g` of `m\times m` matrices over
    `GF(q)` on the points (or subspaces of dimension ``m//2``) isotropic
    w.r.t. the form `F` left invariant by the group `g`.

    The only constraint is that the first ``m//2`` elements of the standard
    basis must generate a totally isotropic w.r.t. `F` subspace; this is the
    case with these groups coming from GAP; namely, `F` has the anti-diagonal
    all-1 matrix.

    INPUT:

    - ``m`` -- the dimension of the underlying vector space

    - ``q`` -- the size of the field

    - ``g`` -- the group acting

    - ``intersection_size`` -- (default: ``None``) if ``None``, build the graph
      on the isotropic points, with adjacency being orthogonality w.r.t. `F`.
      Otherwise, build the graph on the maximal totally isotropic subspaces,
      with adjacency specified by ``intersection_size`` being as given.

    TESTS::

        sage: from sage.graphs.generators.classical_geometries import _polar_graph
        sage: _polar_graph(4, 4, libgap.GeneralUnitaryGroup(4, 2))                                  # needs sage.libs.gap
        Graph on 45 vertices
        sage: _polar_graph(4, 4, libgap.GeneralUnitaryGroup(4, 2), intersection_size=1)             # needs sage.libs.gap
        Graph on 27 vertices
    """
    from sage.libs.gap.libgap import libgap
    from itertools import combinations
    W = libgap.FullRowSpace(libgap.GF(q), m)   # F_q^m
    B = libgap.Elements(libgap.Basis(W))       # the standard basis of W
    V = libgap.Orbit(g, B[0], libgap.OnLines)  # orbit on isotropic points
    gp = libgap.Action(g, V, libgap.OnLines)   # make a permutation group
    s = libgap.Subspace(W, [B[i] for i in range(m//2)])  # a totally isotropic subspace
    # and the points there
    sp = [libgap.Elements(libgap.Basis(x))[0] for x in libgap.Elements(s.Subspaces(1))]
    h = libgap.Set([libgap.Position(V, x) for x in sp])  # indices of the points in s
    L = libgap.Orbit(gp, h, libgap.OnSets)  # orbit on these subspaces
    if intersection_size is None:
        G = Graph()
        for x in L:  # every pair of points in the subspace is adjacent to each other in G
            G.add_edges(combinations(x, 2))
        return G
    else:
        return Graph([L, lambda i, j: libgap.Size(libgap.Intersection(i, j)) == intersection_size],
                     loops=False)


def UnitaryPolarGraph(m, q, algorithm='gap'):
    r"""
    Return the Unitary Polar Graph `U(m,q)`.

    For more information on Unitary Polar graphs, see the `page of Andries
    Brouwer's website <https://www.win.tue.nl/~aeb/graphs/srghub.html>`_.

    INPUT:

    - ``m``, ``q`` -- integers; `q` must be a prime power

    - ``algorithm`` -- string (default: ``'gap'``); if set to ``'gap'`` then the
      computation is carried via GAP library interface, computing totally
      singular subspaces, which is faster for large examples (especially with
      `q>2`). Otherwise it is done directly.

    EXAMPLES::

        sage: # needs sage.libs.gap
        sage: G = graphs.UnitaryPolarGraph(4,2); G
        Unitary Polar Graph U(4, 2); GQ(4, 2): Graph on 45 vertices
        sage: G.is_strongly_regular(parameters=True)
        (45, 12, 3, 3)
        sage: graphs.UnitaryPolarGraph(5,2).is_strongly_regular(parameters=True)
        (165, 36, 3, 9)
        sage: graphs.UnitaryPolarGraph(6,2)     # not tested (long time)
        Unitary Polar Graph U(6, 2): Graph on 693 vertices

    TESTS::

        sage: graphs.UnitaryPolarGraph(4,3, algorithm='gap').is_strongly_regular(parameters=True)   # needs sage.libs.gap
        (280, 36, 8, 4)
        sage: graphs.UnitaryPolarGraph(4,3).is_strongly_regular(parameters=True)                    # needs sage.libs.gap
        (280, 36, 8, 4)
        sage: graphs.UnitaryPolarGraph(4,3, algorithm='foo')
        Traceback (most recent call last):
        ...
        ValueError: unknown algorithm!
    """
    if algorithm == "gap":
        from sage.libs.gap.libgap import libgap
        G = _polar_graph(m, q**2, libgap.GeneralUnitaryGroup(m, q))

    elif algorithm is None:  # slow on large examples
        from sage.schemes.projective.projective_space import ProjectiveSpace
        from sage.modules.free_module_element import free_module_element as vector
        Fq = FiniteField(q**2, 'a')
        PG = map(vector, ProjectiveSpace(m - 1, Fq))

        for v in PG:
            v.set_immutable()

        def P(x, y):
            return sum(x[j] * y[m - 1 - j] ** q for j in range(m)) == 0

        V = [x for x in PG if P(x, x)]
        # bottleneck is here, of course
        G = Graph([V, lambda x, y: P(x, y)], loops=False)
    else:
        raise ValueError("unknown algorithm!")

    G.relabel()
    G.name("Unitary Polar Graph U" + str((m, q)))
    if m == 4:
        G.name(G.name() + '; GQ' + str((q**2, q)))
    if m == 5:
        G.name(G.name() + '; GQ' + str((q**2, q**3)))
    return G


def NonisotropicUnitaryPolarGraph(m, q):
    r"""
    Return the Graph `NU(m,q)`.

    This returns the graph on nonisotropic, with respect to a nondegenerate
    Hermitean form, points of the `(m-1)`-dimensional projective space over
    `F_q`, with points adjacent whenever they lie on a tangent (to the set of
    isotropic points) line.  For more information, see Sect. 9.9 of [BH2012]_
    and series C14 in [Hub1975]_.

    INPUT:

    - ``m``, ``q`` -- integers; `q` must be a prime power

    EXAMPLES::

        sage: g = graphs.NonisotropicUnitaryPolarGraph(5,2); g                          # needs sage.libs.gap
        NU(5, 2): Graph on 176 vertices
        sage: g.is_strongly_regular(parameters=True)                                    # needs sage.libs.gap
        (176, 135, 102, 108)

    TESTS::

        sage: graphs.NonisotropicUnitaryPolarGraph(4,2).is_strongly_regular(parameters=True)        # needs sage.libs.gap
        (40, 27, 18, 18)
        sage: graphs.NonisotropicUnitaryPolarGraph(4,3).is_strongly_regular(parameters=True)  # long time, needs sage.libs.gap
        (540, 224, 88, 96)
        sage: graphs.NonisotropicUnitaryPolarGraph(6,6)
        Traceback (most recent call last):
        ...
        ValueError: q must be a prime power
    """
    p, k = is_prime_power(q, get_data=True)
    if not k:
        raise ValueError('q must be a prime power')
    from sage.libs.gap.libgap import libgap
    from itertools import combinations
    F = libgap.GF(q**2)  # F_{q^2}
    W = libgap.FullRowSpace(F, m)  # F_{q^2}^m
    B = libgap.Elements(libgap.Basis(W))  # the standard basis of W
    if m % 2:
        point = B[(m - 1)/2]
    else:
        if p == 2:
            point = B[m/2] + F.PrimitiveRoot()*B[(m - 2)/2]
        else:
            point = B[(m - 2)/2] + B[m/2]
    g = libgap.GeneralUnitaryGroup(m, q)
    V = libgap.Orbit(g, point, libgap.OnLines)  # orbit on nonisotropic points
    gp = libgap.Action(g, V, libgap.OnLines)  # make a permutation group

    s = libgap.Subspace(W, [point, point + B[0]])  # a tangent line on point

    # and the points there
    sp = [libgap.Elements(libgap.Basis(x))[0] for x in libgap.Elements(s.Subspaces(1))]
    h = libgap.Set([libgap.Position(V, x)
                    for x in libgap.Intersection(V, sp)])  # indices
    L = libgap.Orbit(gp, h, libgap.OnSets)  # orbit on the tangent lines
    G = Graph()
    for x in L:  # every pair of points in the subspace is adjacent to each other in G
        G.add_edges(combinations(x, 2))
    G.relabel()
    G.name("NU" + str((m, q)))
    return G


def UnitaryDualPolarGraph(m, q):
    r"""
    Return the Dual Unitary Polar Graph `U(m,q)`.

    For more information on Unitary Dual Polar graphs, see [BCN1989]_ and
    Sect. 2.3.1 of [Coh1981]_.

    INPUT:

    - ``m``, ``q`` -- integers; `q` must be a prime power

    EXAMPLES:

    The point graph of a generalized quadrangle (see
    :wikipedia:`Generalized_quadrangle`, [PT2009]_) of order (8,4)::

        sage: G = graphs.UnitaryDualPolarGraph(5,2); G  # long time                     # needs sage.libs.gap
        Unitary Dual Polar Graph DU(5, 2); GQ(8, 4): Graph on 297 vertices
        sage: G.is_strongly_regular(parameters=True)    # long time                     # needs sage.libs.gap
        (297, 40, 7, 5)

    Another way to get the  generalized quadrangle of order (2,4)::

        sage: G = graphs.UnitaryDualPolarGraph(4,2); G                                  # needs sage.libs.gap
        Unitary Dual Polar Graph DU(4, 2); GQ(2, 4): Graph on 27 vertices
        sage: G.is_isomorphic(graphs.OrthogonalPolarGraph(6,2,'-'))                     # needs sage.libs.gap
        True

    A bigger graph::

        sage: G = graphs.UnitaryDualPolarGraph(6,2); G   # not tested (long time)
        Unitary Dual Polar Graph DU(6, 2): Graph on 891 vertices
        sage: G.is_distance_regular(parameters=True)     # not tested (long time)
        ([42, 40, 32, None], [None, 1, 5, 21])

    TESTS::

        sage: graphs.UnitaryDualPolarGraph(6,6)                                         # needs sage.libs.gap
        Traceback (most recent call last):
        ...
        GAPError: Error, <subfield> must be a prime or a finite field
    """
    from sage.libs.gap.libgap import libgap
    G = _polar_graph(m, q**2, libgap.GeneralUnitaryGroup(m, q),
                     intersection_size=int((q**(2*(m//2 - 1)) - 1)/(q**2 - 1)))
    G.relabel()
    G.name("Unitary Dual Polar Graph DU" + str((m, q)))
    if m == 4:
        G.name(G.name() + '; GQ' + str((q, q**2)))
    if m == 5:
        G.name(G.name() + '; GQ' + str((q**3, q**2)))
    return G


def SymplecticDualPolarGraph(m, q):
    r"""
    Return the Symplectic Dual Polar Graph `DSp(m,q)`.

    For more information on Symplectic Dual Polar graphs, see [BCN1989]_ and
    Sect. 2.3.1 of [Coh1981]_.

    INPUT:

    - ``m``, ``q`` -- integers; `q` must be a prime power, and `m` must be even

    EXAMPLES::

        sage: G = graphs.SymplecticDualPolarGraph(6,3); G       # not tested (long time)
        Symplectic Dual Polar Graph DSp(6, 3): Graph on 1120 vertices
        sage: G.is_distance_regular(parameters=True)            # not tested (long time)
        ([39, 36, 27, None], [None, 1, 4, 13])

    TESTS::

        sage: G = graphs.SymplecticDualPolarGraph(6,2); G                               # needs sage.libs.gap
        Symplectic Dual Polar Graph DSp(6, 2): Graph on 135 vertices
        sage: G.is_distance_regular(parameters=True)                                    # needs sage.libs.gap
        ([14, 12, 8, None], [None, 1, 3, 7])
        sage: graphs.SymplecticDualPolarGraph(6,6)                                      # needs sage.libs.gap
        Traceback (most recent call last):
        ...
        GAPError: Error, <subfield> must be a prime or a finite field
    """
    from sage.libs.gap.libgap import libgap
    G = _polar_graph(m, q, libgap.SymplecticGroup(m, q),
                     intersection_size=int((q**(m/2 - 1) - 1)/(q - 1)))

    G.relabel()
    G.name("Symplectic Dual Polar Graph DSp" + str((m, q)))
    if m == 4:
        G.name(G.name() + '; GQ' + str((q, q)))
    return G


def TaylorTwographDescendantSRG(q, clique_partition=False):
    r"""
    Return the descendant graph of the Taylor's two-graph for `U_3(q)`, `q` odd.

    This is a strongly regular graph with parameters
    `(v,k,\lambda,\mu)=(q^3, (q^2+1)(q-1)/2, (q-1)^3/4-1, (q^2+1)(q-1)/4)`
    obtained as a two-graph descendant of the
    :func:`Taylor's two-graph <sage.combinat.designs.twographs.taylor_twograph>` `T`.
    This graph admits a partition into cliques of size `q`, which are useful in
    :func:`~sage.graphs.graph_generators.GraphGenerators.TaylorTwographSRG`,
    a strongly regular graph on `q^3+1` vertices in the
    Seidel switching class of `T`, for which we need `(q^2+1)/2` cliques.
    The cliques are the `q^2` lines on `v_0` of the projective plane containing
    the unital for `U_3(q)`, and intersecting the unital (i.e. the vertices of
    the graph and the point we remove) in `q+1` points. This is all taken from
    §7E of [BL1984]_.

    INPUT:

    - ``q`` -- a power of an odd prime number

    - ``clique_partition`` -- boolean (default: ``False``); when set to
      ``True``, return `q^2-1` cliques of size `q` with empty pairwise
      intersection. (Removing all of them leaves a clique, too), and the point
      removed from the unital.

    EXAMPLES::

        sage: # needs sage.rings.finite_rings
        sage: g = graphs.TaylorTwographDescendantSRG(3); g
        Taylor two-graph descendant SRG: Graph on 27 vertices
        sage: g.is_strongly_regular(parameters=True)
        (27, 10, 1, 5)
        sage: from sage.combinat.designs.twographs import taylor_twograph
        sage: T = taylor_twograph(3)                            # long time
        sage: g.is_isomorphic(T.descendant(T.ground_set()[1]))  # long time
        True
        sage: g = graphs.TaylorTwographDescendantSRG(5)         # not tested (long time)
        sage: g.is_strongly_regular(parameters=True)            # not tested (long time)
        (125, 52, 15, 26)

    TESTS::

        sage: # needs sage.rings.finite_rings
        sage: g,l,_ = graphs.TaylorTwographDescendantSRG(3, clique_partition=True)
        sage: all(g.is_clique(x) for x in l)
        True
        sage: graphs.TaylorTwographDescendantSRG(4)
        Traceback (most recent call last):
        ...
        ValueError: q must be an odd prime power
        sage: graphs.TaylorTwographDescendantSRG(6)
        Traceback (most recent call last):
        ...
        ValueError: q must be an odd prime power
    """
    p, k = is_prime_power(q, get_data=True)
    if not k or p == 2:
        raise ValueError('q must be an odd prime power')
    from sage.schemes.projective.projective_space import ProjectiveSpace
    from sage.rings.finite_rings.integer_mod import mod
    Fq = FiniteField(q**2, 'a')
    PG = map(tuple, ProjectiveSpace(2, Fq))

    def S(x, y):
        return sum(x[j] * y[2 - j] ** q for j in range(3))

    V = [x for x in PG if S(x, x) == 0]  # the points of the unital
    v0 = V[0]
    V.remove(v0)
    if mod(q, 4) == 1:
        G = Graph([V, lambda y, z: not (S(v0, y)*S(y, z)*S(z, v0)).is_square()], loops=False)
    else:
        G = Graph([V, lambda y, z: (S(v0, y)*S(y, z)*S(z, v0)).is_square()], loops=False)
    G.name("Taylor two-graph descendant SRG")
    if clique_partition:
        lines = [[t for t in V if t[0] + z * t[1] == 0]
                 for z in Fq if z]
        return (G, lines, v0)
    else:
        return G


def TaylorTwographSRG(q):
    r"""
    Return a strongly regular graph from the Taylor's two-graph for `U_3(q)`,
    `q` odd

    This is a strongly regular graph with parameters
    `(v,k,\lambda,\mu)=(q^3+1, q(q^2+1)/2, (q^2+3)(q-1)/4, (q^2+1)(q+1)/4)`
    in the Seidel switching class of
    :func:`Taylor two-graph <sage.combinat.designs.twographs.taylor_twograph>`.
    Details are in §7E of [BL1984]_.

    INPUT:

    - ``q`` -- a power of an odd prime number

    .. SEEALSO::

        * :meth:`~sage.graphs.graph_generators.GraphGenerators.TaylorTwographDescendantSRG`

    EXAMPLES::

        sage: t = graphs.TaylorTwographSRG(3); t                                        # needs sage.rings.finite_rings
        Taylor two-graph SRG: Graph on 28 vertices
        sage: t.is_strongly_regular(parameters=True)                                    # needs sage.rings.finite_rings
        (28, 15, 6, 10)
    """
    G, l, v0 = TaylorTwographDescendantSRG(q, clique_partition=True)
    G.add_vertex(v0)
    G.seidel_switching(sum(l[:(q**2 + 1)/2], []))
    G.name("Taylor two-graph SRG")
    return G


def AhrensSzekeresGeneralizedQuadrangleGraph(q, dual=False):
    r"""
    Return the collinearity graph of the generalized quadrangle `AS(q)`, or of
    its dual

    Let `q` be an odd prime power.  `AS(q)` is a generalized quadrangle
    (:wikipedia:`Generalized_quadrangle`) of
    order `(q-1,q+1)`, see 3.1.5 in [PT2009]_. Its points are elements
    of `F_q^3`, and lines are sets of size `q` of the form

    * `\{ (\sigma, a, b) \mid \sigma\in F_q \}`
    * `\{ (a, \sigma, b) \mid \sigma\in F_q \}`
    * `\{ (c \sigma^2 - b \sigma + a, -2 c \sigma + b, \sigma) \mid \sigma\in F_q \}`,

    where `a`, `b`, `c` are arbitrary elements of `F_q`.

    INPUT:

    - ``q`` -- a power of an odd prime number

    - ``dual`` -- boolean (default: ``False``); whether to return the
      collinearity graph of `AS(q)` or of the dual `AS(q)` (when ``True``)

    EXAMPLES::

        sage: g = graphs.AhrensSzekeresGeneralizedQuadrangleGraph(5); g
        AS(5); GQ(4, 6): Graph on 125 vertices
        sage: g.is_strongly_regular(parameters=True)
        (125, 28, 3, 7)
        sage: g = graphs.AhrensSzekeresGeneralizedQuadrangleGraph(5, dual=True); g
        AS(5)*; GQ(6, 4): Graph on 175 vertices
        sage: g.is_strongly_regular(parameters=True)
        (175, 30, 5, 5)
    """
    from sage.combinat.designs.incidence_structures import IncidenceStructure
    p, k = is_prime_power(q, get_data=True)
    if not k or p == 2:
        raise ValueError('q must be an odd prime power')
    F = FiniteField(q, 'a')
    L = []
    for a in F:
        for b in F:
            L.append(tuple((s, a, b) for s in F))
            L.append(tuple((a, s, b) for s in F))
            for c in F:
                L.append(tuple((c*s**2 - b*s + a, -2*c*s + b, s) for s in F))
    if dual:
        G = IncidenceStructure(L).intersection_graph()
        G.name('AS(' + str(q) + ')*; GQ' + str((q + 1, q - 1)))
    else:
        G = IncidenceStructure(L).dual().intersection_graph()
        G.name('AS(' + str(q) + '); GQ' + str((q - 1, q + 1)))
    return G


def T2starGeneralizedQuadrangleGraph(q, dual=False, hyperoval=None, field=None, check_hyperoval=True):
    r"""
    Return the collinearity graph of the generalized quadrangle `T_2^*(q)`, or
    of its dual

    Let `q=2^k` and `\Theta=PG(3,q)`.  `T_2^*(q)` is a generalized quadrangle
    (:wikipedia:`Generalized_quadrangle`)
    of order `(q-1,q+1)`, see 3.1.3 in [PT2009]_. Fix a plane `\Pi \subset
    \Theta` and a
    `hyperoval <http://en.wikipedia.org/wiki/Oval_(projective_plane)#Even_q>`__
    `O \subset \Pi`. The points of `T_2^*(q):=T_2^*(O)` are the points of
    `\Theta` outside `\Pi`, and the lines are the lines of `\Theta` outside
    `\Pi` that meet `\Pi` in a point of `O`.

    INPUT:

    - ``q`` -- a power of two

    - ``dual`` -- boolean (default: ``False``); whether to return the graph of
      `T_2^*(O)` or of the dual `T_2^*(O)` (when ``True``)

    - ``hyperoval`` -- a hyperoval (i.e. a complete 2-arc; a set of points in
      the plane meeting every line in 0 or 2 points) in the plane of points with
      0th coordinate 0 in `PG(3,q)` over the field ``field``. Each point of
      ``hyperoval`` must be a length 4 vector over ``field`` with 1st non-0
      coordinate equal to 1. By default, ``hyperoval`` and ``field`` are not
      specified, and constructed on the fly. In particular, ``hyperoval`` we
      build is the classical one, i.e. a conic with the point of intersection of
      its tangent lines.

    - ``field`` -- an instance of a finite field of order `q`, must be provided
      if ``hyperoval`` is provided

    - ``check_hyperoval`` -- boolean (default: ``True``); whether to check
      ``hyperoval`` for correctness or not

    EXAMPLES:

    using the built-in construction::

        sage: # needs sage.combinat sage.rings.finite_rings
        sage: g = graphs.T2starGeneralizedQuadrangleGraph(4); g
        T2*(O,4); GQ(3, 5): Graph on 64 vertices
        sage: g.is_strongly_regular(parameters=True)
        (64, 18, 2, 6)
        sage: g = graphs.T2starGeneralizedQuadrangleGraph(4, dual=True); g
        T2*(O,4)*; GQ(5, 3): Graph on 96 vertices
        sage: g.is_strongly_regular(parameters=True)
        (96, 20, 4, 4)

    supplying your own hyperoval::

        sage: # needs sage.combinat sage.rings.finite_rings
        sage: F = GF(4,'b')
        sage: O = [vector(F,(0,0,0,1)),vector(F,(0,0,1,0))] + [vector(F, (0,1,x^2,x))
        ....:                                                  for x in F]
        sage: g = graphs.T2starGeneralizedQuadrangleGraph(4, hyperoval=O, field=F); g
        T2*(O,4); GQ(3, 5): Graph on 64 vertices
        sage: g.is_strongly_regular(parameters=True)
        (64, 18, 2, 6)

    TESTS::

        sage: # needs sage.combinat sage.rings.finite_rings
        sage: F = GF(4,'b')  # repeating a point...
        sage: O = [vector(F,(0,1,0,0)),vector(F,(0,0,1,0))]+[vector(F, (0,1,x^2,x)) for x in F]
        sage: graphs.T2starGeneralizedQuadrangleGraph(4, hyperoval=O, field=F)
        Traceback (most recent call last):
        ...
        RuntimeError: incorrect hyperoval size
        sage: O = [vector(F,(0,1,1,0)),vector(F,(0,0,1,0))]+[vector(F, (0,1,x^2,x)) for x in F]
        sage: graphs.T2starGeneralizedQuadrangleGraph(4, hyperoval=O, field=F)
        Traceback (most recent call last):
        ...
        RuntimeError: incorrect hyperoval
    """
    from sage.combinat.designs.incidence_structures import IncidenceStructure
    from sage.combinat.designs.block_design import ProjectiveGeometryDesign as PG

    p, k = is_prime_power(q, get_data=True)
    if not k or p != 2:
        raise ValueError('q must be a power of 2')
    if field is None:
        F = FiniteField(q, 'a')
    else:
        F = field

    Theta = PG(3, 1, F, point_coordinates=1)
    Pi = set(x for x in Theta.ground_set() if x[0] == F.zero())
    if hyperoval is None:
        HO = set(x for x in Pi
                 if (x[1] + x[2] * x[3] == 0) or
                    (x[1] == 1 and x[2] == x[3] == 0))
    else:
        for v in hyperoval:
            v.set_immutable()

        HO = set(hyperoval)
        if check_hyperoval:
            if len(HO) != q + 2:
                raise RuntimeError("incorrect hyperoval size")
            for L in Theta.blocks():
                if set(L).issubset(Pi):
                    if len(HO.intersection(L)) not in [0, 2]:
                        raise RuntimeError("incorrect hyperoval")

    L = [[y for y in z if y not in HO]
         for z in [x for x in Theta.blocks() if len(HO.intersection(x)) == 1]]

    if dual:
        G = IncidenceStructure(L).intersection_graph()
        G.name('T2*(O,' + str(q) + ')*; GQ' + str((q + 1, q - 1)))
    else:
        G = IncidenceStructure(L).dual().intersection_graph()
        G.name('T2*(O,' + str(q) + '); GQ' + str((q - 1, q + 1)))
    return G


def HaemersGraph(q, hyperoval=None, hyperoval_matching=None, field=None, check_hyperoval=True):
    r"""
    Return the Haemers graph obtained from `T_2^*(q)^*`.

    Let `q` be a power of 2. In Sect. 8.A of [BL1984]_ one finds a construction
    of a strongly regular graph with parameters `(q^2(q+2),q^2+q-1,q-2,q)` from
    the graph of `T_2^*(q)^*`, constructed by
    :func:`~sage.graphs.graph_generators.GraphGenerators.T2starGeneralizedQuadrangleGraph`,
    by redefining adjacencies in the way specified by an arbitrary
    ``hyperoval_matching`` of the points (i.e. partitioning into size two parts)
    of ``hyperoval`` defining `T_2^*(q)^*`.

    While [BL1984]_ gives the construction in geometric terms, it can be
    formulated, and is implemented, in graph-theoretic ones, of re-adjusting the
    edges. Namely, `G=T_2^*(q)^*` has a partition into `q+2` independent sets
    `I_k` of size `q^2` each. Each vertex in `I_j` is adjacent to `q` vertices
    from `I_k`. Each `I_k` is paired to some `I_{k'}`, according to
    ``hyperoval_matching``. One adds edges `(s,t)` for `s,t \in I_k` whenever
    `s` and `t` are adjacent to some `u \in I_{k'}`, and removes all the edges
    between `I_k` and `I_{k'}`.

    INPUT:

    - ``q`` -- a power of two

    - ``hyperoval_matching`` -- if ``None`` (default), pair each `i`-th point of
      ``hyperoval`` with `(i+1)`-th. Otherwise, specifies the pairing
      in the format `((i_1,i'_1),(i_2,i'_2),...)`.

    - ``hyperoval`` -- a hyperoval defining `T_2^*(q)^*`. If ``None`` (default),
      the classical hyperoval obtained from a conic is used. See the
      documentation of
      :func:`~sage.graphs.graph_generators.GraphGenerators.T2starGeneralizedQuadrangleGraph`,
      for more information.

    - ``field`` -- an instance of a finite field of order `q`, must be provided
      if ``hyperoval`` is provided

    - ``check_hyperoval`` -- boolean (default: ``True``); whether to check
      ``hyperoval`` for correctness or not

    EXAMPLES:

    using the built-in constructions::

        sage: # needs sage.combinat sage.rings.finite_rings
        sage: g = graphs.HaemersGraph(4); g
        Haemers(4): Graph on 96 vertices
        sage: g.is_strongly_regular(parameters=True)
        (96, 19, 2, 4)

    supplying your own hyperoval_matching::

        sage: # needs sage.combinat sage.rings.finite_rings
        sage: g = graphs.HaemersGraph(4, hyperoval_matching=((0,5),(1,4),(2,3))); g
        Haemers(4): Graph on 96 vertices
        sage: g.is_strongly_regular(parameters=True)
        (96, 19, 2, 4)

    TESTS::

        sage: # needs sage.combinat sage.rings.finite_rings
        sage: F = GF(4,'b')  # repeating a point...
        sage: O = [vector(F,(0,1,0,0)),vector(F,(0,0,1,0))]+[vector(F, (0,1,x^2,x)) for x in F]
        sage: graphs.HaemersGraph(4, hyperoval=O, field=F)
        Traceback (most recent call last):
        ...
        RuntimeError: incorrect hyperoval size
        sage: O = [vector(F,(0,1,1,0)),vector(F,(0,0,1,0))]+[vector(F, (0,1,x^2,x)) for x in F]
        sage: graphs.HaemersGraph(4, hyperoval=O, field=F)
        Traceback (most recent call last):
        ...
        RuntimeError: incorrect hyperoval

        sage: g = graphs.HaemersGraph(8); g             # not tested (long time)        # needs sage.rings.finite_rings
        Haemers(8): Graph on 640 vertices
        sage: g.is_strongly_regular(parameters=True)    # not tested (long time)        # needs sage.rings.finite_rings
        (640, 71, 6, 8)
    """
    from sage.modules.free_module_element import free_module_element as vector
    from sage.rings.finite_rings.finite_field_constructor import GF
    from itertools import combinations

    p, k = is_prime_power(q, get_data=True)
    if not k or p != 2:
        raise ValueError('q must be a power of 2')

    if hyperoval_matching is None:
        hyperoval_matching = [(2 * K + 1, 2 * K) for K in range(1 + q // 2)]
    if field is None:
        F = GF(q, 'a')
    else:
        F = field

    # for q=8, 95% of CPU time taken by this function is spent in the following call
    G = T2starGeneralizedQuadrangleGraph(q, field=F, dual=True,
                                         hyperoval=hyperoval,
                                         check_hyperoval=check_hyperoval)

    def normalize(v):  # make sure the 1st non-0 coordinate is 1.
        d = next(x for x in v if x != F.zero())
        return vector([x / d for x in v])

    # build the partition into independent sets
    P = [tuple(normalize(v[0] - v[1])) for v in G.vertices(sort=True)]
    Pi_to_int = {Pi: i for i, Pi in enumerate(set(P))}
    I_ks = {x: [] for x in range(q + 2)}  # the partition into I_k's
    for i, Pi in enumerate(P):
        I_ks[Pi_to_int[tuple(Pi)]].append(i)

    # perform the adjustment of the edges, as described.
    G.relabel(range(G.order()))
    cliques = []
    for i, j in hyperoval_matching:
        Pij = set(I_ks[i] + I_ks[j])
        for v in Pij:
            cliques.append(Pij.intersection(G.neighbors(v)))
        G.delete_edges(G.edge_boundary(I_ks[i], I_ks[j]))  # edges on (I_i,I_j)
    G.add_edges(e for c in cliques for e in combinations(c, 2))
    G.name('Haemers(' + str(q) + ')')
    return G


def CossidentePenttilaGraph(q):
    r"""
    Return the Cossidente-Penttila
    `((q^3+1)(q+1)/2,(q^2+1)(q-1)/2,(q-3)/2,(q-1)^2/2)`-strongly regular graph

    For each odd prime power `q`, one can partition the points of the
    `O_6^-(q)`-generalized quadrangle `GQ(q,q^2)` into two parts, so that on any
    of them the induced subgraph of the point graph of the GQ has parameters as
    above [CP2005]_.

    Directly following the construction in [CP2005]_ is not efficient, as one
    then needs to construct the dual `GQ(q^2,q)`. Thus we describe here a more
    efficient approach that we came up with, following a suggestion by
    T.Penttila. Namely, this partition is invariant under the subgroup
    `H=\Omega_3(q^2)<O_6^-(q)`. We build the appropriate `H`, which leaves the
    form `B(X,Y,Z)=XY+Z^2` invariant, and pick up two orbits of `H` on the
    `F_q`-points. One them is `B`-isotropic, and we take the representative
    `(1:0:0)`. The other one corresponds to the points of `PG(2,q^2)` that have
    all the lines on them either missing the conic specified by `B`, or
    intersecting the conic in two points. We take `(1:1:e)` as the
    representative. It suffices to pick `e` so that `e^2+1` is not a square in
    `F_{q^2}`. Indeed, The conic can be viewed as the union of `\{(0:1:0)\}` and
    `\{(1:-t^2:t) | t \in F_{q^2}\}`.  The coefficients of a generic line on
    `(1:1:e)` are `[1:-1-eb:b]`, for `-1\neq eb`.  Thus, to make sure the
    intersection with the conic is always even, we need that the discriminant of
    `1+(1+eb)t^2+tb=0` never vanishes, and this is if and only if `e^2+1` is not
    a square. Further, we need to adjust `B`, by multiplying it by appropriately
    chosen `\nu`, so that `(1:1:e)` becomes isotropic under the relative trace
    norm `\nu B(X,Y,Z)+(\nu B(X,Y,Z))^q`. The latter is used then to define the
    graph.

    INPUT:

    - ``q`` -- an odd prime power

    EXAMPLES:

    For `q=3` one gets Sims-Gewirtz graph. ::

        sage: G = graphs.CossidentePenttilaGraph(3)     # optional - gap_package_grape
        sage: G.is_strongly_regular(parameters=True)    # optional - gap_package_grape
        (56, 10, 0, 2)

    For `q>3` one gets new graphs. ::

        sage: G = graphs.CossidentePenttilaGraph(5)     # optional - gap_package_grape
        sage: G.is_strongly_regular(parameters=True)    # optional - gap_package_grape
        (378, 52, 1, 8)

    TESTS::

        sage: G = graphs.CossidentePenttilaGraph(7)     # optional - gap_package_grape, long time
        sage: G.is_strongly_regular(parameters=True)    # optional - gap_package_grape, long time
        (1376, 150, 2, 18)
        sage: graphs.CossidentePenttilaGraph(2)
        Traceback (most recent call last):
        ...
        ValueError: q(=2) must be an odd prime power
    """
    p, k = is_prime_power(q, get_data=True)
    if not k or p == 2:
        raise ValueError('q(={}) must be an odd prime power'.format(q))

    from sage.features.gap import GapPackage
    GapPackage("grape", spkg='gap_packages').require()

    from sage.libs.gap.libgap import libgap
    adj_list = libgap.function_factory("""function(q)
        local z, e, so, G, nu, G1, G0, B, T, s, O1, O2, x, sqo;
        LoadPackage("grape");
        G0:=SO(3,q^2);
        so:=GeneratorsOfGroup(G0);
        G1:=Group(Comm(so[1],so[2]),Comm(so[1],so[3]),Comm(so[2],so[3]));
        B:=InvariantBilinearForm(G0).matrix;
        z:=Z(q^2); e:=z; sqo:=(q^2-1)/2;
        if IsInt(sqo/Order(e^2+z^0)) then
            e:=z^First([2..q^2-2], x-> not IsInt(sqo/Order(z^(2*x)+z^0)));
        fi;
        nu:=z^First([0..q^2-2], x->z^x*(e^2+z^0)+(z^x*(e^2+z^0))^q=0*z);
        T:=function(x)
            local r;
            r:=nu*x*B*x;
            return r+r^q;
        end;
        s:=Group([Z(q)*IdentityMat(3,GF(q))]);
        O1:=Orbit(G1, Set(Orbit(s,z^0*[1,0,0])), OnSets);
        O2:=Orbit(G1, Set(Orbit(s,z^0*[1,1,e])), OnSets);
        G:=Graph(G1,Concatenation(O1,O2),OnSets,
            function(x,y) return x<>y and 0*z=T(x[1]+y[1]); end);
        return List([1..OrderGraph(G)],x->Adjacency(G,x));
        end;""")

    adj = adj_list(q)  # for each vertex, we get the list of vertices it is adjacent to
    G = Graph(((i, int(j - 1))
               for i, ni in enumerate(adj) for j in ni),
              format='list_of_edges', multiedges=False)
    G.name('CossidentePenttila(' + str(q) + ')')
    return G


def Nowhere0WordsTwoWeightCodeGraph(q, hyperoval=None, field=None, check_hyperoval=True):
    r"""
    Return the subgraph of nowhere 0 words from two-weight code of projective
    plane hyperoval.

    Let `q=2^k` and `\Pi=PG(2,q)`.  Fix a
    `hyperoval <http://en.wikipedia.org/wiki/Oval_(projective_plane)#Even_q>`__
    `O \subset \Pi`. Let `V=F_q^3` and `C` the two-weight 3-dimensional linear
    code over `F_q` with words `c(v)` obtained from `v\in V` by computing

    .. MATH::

        c(v)=(\langle v,o_1 \rangle,...,\langle v,o_{q+2} \rangle), o_j \in O.

    `C` contains `q(q-1)^2/2` words without 0 entries. The subgraph of the
    strongly regular graph of `C` induced on the latter words is also strongly
    regular, assuming `q>4`. This is a construction due to A.E.Brouwer
    [Bro2016]_, and leads to graphs with parameters also given by a construction
    in [HHL2009]_.  According to [Bro2016]_, these two constructions are likely
    to produce isomorphic graphs.

    INPUT:

    - ``q`` -- a power of two

    - ``hyperoval`` -- a hyperoval (i.e. a complete 2-arc; a set of points in
      the plane meeting every line in 0 or 2 points) in `PG(2,q)` over the field
      ``field``.  Each point of ``hyperoval`` must be a length 3 vector over
      ``field`` with 1st non-0 coordinate equal to 1. By default, ``hyperoval``
      and ``field`` are not specified, and constructed on the fly. In
      particular, ``hyperoval`` we build is the classical one, i.e. a conic with
      the point of intersection of its tangent lines.

    - ``field`` -- an instance of a finite field of order `q`; must be provided
      if ``hyperoval`` is provided

    - ``check_hyperoval`` -- boolean (default: ``True``); whether to check
      ``hyperoval`` for correctness or not

    .. SEEALSO::

        - :func:`~sage.graphs.strongly_regular_db.is_nowhere0_twoweight`

    EXAMPLES:

    using the built-in construction::

        sage: # needs sage.combinat sage.rings.finite_rings
        sage: g = graphs.Nowhere0WordsTwoWeightCodeGraph(8); g
        Nowhere0WordsTwoWeightCodeGraph(8): Graph on 196 vertices
        sage: g.is_strongly_regular(parameters=True)
        (196, 60, 14, 20)
        sage: g = graphs.Nowhere0WordsTwoWeightCodeGraph(16)  # not tested (long time)
        sage: g.is_strongly_regular(parameters=True)          # not tested (long time)
        (1800, 728, 268, 312)

    supplying your own hyperoval::

        sage: # needs sage.combinat sage.rings.finite_rings
        sage: F = GF(8)
        sage: O = [vector(F,(0,0,1)),vector(F,(0,1,0))] + [vector(F, (1,x^2,x))
        ....:                                              for x in F]
        sage: g = graphs.Nowhere0WordsTwoWeightCodeGraph(8,hyperoval=O,field=F); g
        Nowhere0WordsTwoWeightCodeGraph(8): Graph on 196 vertices
        sage: g.is_strongly_regular(parameters=True)
        (196, 60, 14, 20)

    TESTS::

        sage: # needs sage.combinat sage.rings.finite_rings
        sage: F = GF(8)  # repeating a point...
        sage: O = [vector(F,(1,0,0)),vector(F,(0,1,0))]+[vector(F, (1,x^2,x)) for x in F]
        sage: graphs.Nowhere0WordsTwoWeightCodeGraph(8,hyperoval=O,field=F)
        Traceback (most recent call last):
        ...
        RuntimeError: incorrect hyperoval size
        sage: O = [vector(F,(1,1,0)),vector(F,(0,1,0))]+[vector(F, (1,x^2,x)) for x in F]
        sage: graphs.Nowhere0WordsTwoWeightCodeGraph(8,hyperoval=O,field=F)
        Traceback (most recent call last):
        ...
        RuntimeError: incorrect hyperoval
    """
    from sage.combinat.designs.block_design import ProjectiveGeometryDesign as PG
    from sage.matrix.constructor import matrix

    p, k = is_prime_power(q, get_data=True)
    if not k or p != 2:
        raise ValueError('q must be a power of 2')
    if k < 3:
        raise ValueError('q must be a at least 8')
    if field is None:
        F = FiniteField(q, 'a')
    else:
        F = field

    Theta = PG(2, 1, F, point_coordinates=1)
    Pi = Theta.ground_set()
    if hyperoval is None:
        hyperoval = [x for x in Pi
                     if (x[0] + x[1] * x[2] == 0) or
                        (x[0] == 1 and x[1] == x[2] == 0)]
    else:
        for v in hyperoval:
            v.set_immutable()

        if check_hyperoval:
            HO = set(hyperoval)
            if len(HO) != q + 2:
                raise RuntimeError("incorrect hyperoval size")
            for L in Theta.blocks():
                if set(L).issubset(Pi):
                    if len(HO.intersection(L)) not in [0, 2]:
                        raise RuntimeError("incorrect hyperoval")
    M = matrix(hyperoval)
    F_0 = F.zero()
    C = [p for p in [M*x for x in F**3] if F_0 not in p]

    for x in C:
        x.set_immutable()
    G = Graph([C, lambda x, y: F.zero() not in x + y])
    G.name('Nowhere0WordsTwoWeightCodeGraph(' + str(q) + ')')
    G.relabel()
    return G


def OrthogonalDualPolarGraph(e, d, q):
    r"""
    Return the dual polar graph on `GO^e(n,q)` of diameter `d`.

    The value of `n` is determined by `d` and `e`.

    The graph is distance-regular with classical parameters `(d, q, 0, q^e)`.

    INPUT:

    - ``e`` -- integer; type of the orthogonal polar space to consider;
      must be `-1, 0` or  `1`

    - ``d`` -- integer; diameter of the graph

    - ``q`` -- integer; prime power; order of the finite field over which to
      build the polar space

    EXAMPLES::

        sage: # needs sage.libs.gap
        sage: G = graphs.OrthogonalDualPolarGraph(1,3,2)
        sage: G.is_distance_regular(True)
        ([7, 6, 4, None], [None, 1, 3, 7])
        sage: G = graphs.OrthogonalDualPolarGraph(0,3,3)        # long time
        sage: G.is_distance_regular(True)                       # long time
        ([39, 36, 27, None], [None, 1, 4, 13])
        sage: G.order()                                         # long time
        1120

    REFERENCES:

    See [BCN1989]_ pp. 274-279 or [VDKT2016]_ p. 22.

    TESTS::

        sage: # needs sage.libs.gap
        sage: G = graphs.OrthogonalDualPolarGraph(0,3,2)
        sage: G.is_distance_regular(True)
        ([14, 12, 8, None], [None, 1, 3, 7])
        sage: G = graphs.OrthogonalDualPolarGraph(-1,3,2)       # long time
        sage: G.is_distance_regular(True)                       # long time
        ([28, 24, 16, None], [None, 1, 3, 7])
        sage: G = graphs.OrthogonalDualPolarGraph(1,3,4)
        sage: G.is_distance_regular(True)
        ([21, 20, 16, None], [None, 1, 5, 21])
        sage: G = graphs.OrthogonalDualPolarGraph(1,4,2)
        sage: G.is_distance_regular(True)
        ([15, 14, 12, 8, None], [None, 1, 3, 7, 15])
    """
    from sage.libs.gap.libgap import libgap
    from sage.matrix.constructor import Matrix
    from sage.modules.free_module import VectorSpace
    from sage.rings.finite_rings.finite_field_constructor import GF
    import itertools

    def hashable(v):
        v.set_immutable()
        return v

    if e not in {0, 1, -1}:
        raise ValueError("e must by 0, +1 or -1")

    m = 2*d + 1 - e

    group = libgap.GeneralOrthogonalGroup(e, m, q)
    M = Matrix(libgap.InvariantQuadraticForm(group)["matrix"])
    # Q(x) = xMx is our quadratic form

    # we need to find a totally isotropic subspace of dimension d
    K = M.kernel()
    isotropicBasis = list(K.basis())

    # extend K to a maximal isotropic subspace
    if K.dimension() < d:
        V = VectorSpace(GF(q), m)

        # get all projective points not in K
        candidates = set(map(hashable, [P.basis()[0] for P in V.subspaces(1)]))
        hashableK = map(hashable, [P.basis()[0] for P in K.subspaces(1)])
        candidates = candidates.difference(hashableK)

        nonZeroScalars = [x for x in GF(q) if not x.is_zero()]
        while K.dimension() < d:
            found = False
            while not found:
                v = candidates.pop()
                if v*M*v == 0:
                    # found another isotropic point
                    # check if we can add it to K
                    found = True
                    for w in isotropicBasis:
                        if w*M*v + v*M*w != 0:
                            found = False
                            break
            # here we found a valid point
            isotropicBasis.append(v)

            # remove new points of K
            newVectors = map(hashable,
                             [k + s*v for k in K for s in nonZeroScalars])
            candidates.difference(newVectors)
            K = V.span(isotropicBasis)

        # here K is a totally isotropic subspace of dimension d
        isotropicBasis = list(K.basis())

    W = libgap.FullRowSpace(libgap.GF(q), m)
    isoS = libgap.Subspace(W, isotropicBasis)  # gap version of K

    allIsoPoints = libgap.Orbit(group, isotropicBasis[0], libgap.OnLines)
    permutation = libgap.Action(group, allIsoPoints, libgap.OnLines)
    # this is the permutation group generated by GO^e(n,q) acting on
    # projective isotropic points

    # convert isoS into a list of ints representing the projective points
    isoSPoints = [libgap.Elements(libgap.Basis(x))[0]
                  for x in libgap.Elements(isoS.Subspaces(1))]
    isoSPointsInt = libgap.Set([libgap.Position(allIsoPoints, x)
                                for x in isoSPoints])

    # all isotropic subspaces of dimension d
    allIsoSubspaces = libgap.Orbit(permutation, isoSPointsInt, libgap.OnSets)

    # number of projective points in a (d-1)-subspace
    intersection_size = (q**(d-1) - 1) // (q-1)

    edges = []
    n = len(allIsoSubspaces)
    for i, j in itertools.combinations(range(n), 2):
        if libgap.Size(libgap.Intersection(allIsoSubspaces[i],
                                           allIsoSubspaces[j])) \
                                           == intersection_size:
            edges.append((i, j))

    G = Graph(edges, format='list_of_edges')
    G.name("Dual Polar Graph on Orthogonal group (%d, %d, %d)" % (e, m, q))
    return G



# sage_setup: distribution = sagemath-graphs
# sage.doctest: needs sage.modules
r"""
Elements of posets, lattices, semilattices, etc.
"""
# ****************************************************************************
#       Copyright (C) 2008 Peter Jipsen <jipsen@chapman.edu>,
#                          Franco Saliola <saliola@gmail.com>
#
#  Distributed under the terms of the GNU General Public License (GPL)
#
#    This code is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
#    General Public License for more details.
#
#  The full text of the GPL is available at:
#
#                  https://www.gnu.org/licenses/
# ****************************************************************************
from sage.structure.element import Element
from sage.structure.element import have_same_parent


class PosetElement(Element):

    def __init__(self, poset, element, vertex) -> None:
        r"""
        Establish the parent-child relationship between ``poset``
        and ``element``, where ``element`` is associated to the
        vertex ``vertex`` of the Hasse diagram of the poset.

        INPUT:

        - ``poset`` -- a poset object

        - ``element`` -- any object

        - ``vertex`` -- a vertex of the Hasse diagram of the poset

        TESTS::

            sage: from sage.combinat.posets.elements import PosetElement
            sage: P = Poset([[1,2],[4],[3],[4],[]], facade = False)
            sage: e = P(0)
            sage: e.parent() is P
            True
            sage: TestSuite(e).run()
        """
        Element.__init__(self, poset)
        if isinstance(element, self.parent().element_class):
            self.element = element.element
        else:
            self.element = element
        self.vertex = vertex

    def __hash__(self):
        r"""
        TESTS::

            sage: P = Poset([[1,2],[4],[3],[4],[]], facade = False)
            sage: e = P(0)
            sage: hash(e)
            0
        """
        return hash(self.element)

    def _repr_(self):
        """
        TESTS::

            sage: Poset([[1,2],[4],[3],[4],[]], facade = False)(0)._repr_()
            '0'
        """
        return "%s" % str(self.element)

    def _latex_(self):
        r"""
        Return the latex code of the poset element.

        EXAMPLES::

            sage: m = matrix(2, [1,2,3,4])
            sage: m.set_immutable()
            sage: P = Poset(([m],[]), facade=False)
            sage: [e] = P
            sage: type(e)
            <class 'sage.combinat.posets.posets.FinitePoset_with_category.element_class'>
            sage: latex(e)                 #indirect doctest
            \left(\begin{array}{rr}
            1 & 2 \\
            3 & 4
            \end{array}\right)
        """
        from sage.misc.latex import latex
        return latex(self.element)

    def __eq__(self, other):
        """
        TESTS::

            sage: P = Poset([["a","b"],["d"],["c"],["d"],[]], facade = False)
            sage: Q = Poset([["a","b"],["d"],["c"],[],[]], facade = False)
            sage: P(0).__eq__(P(4))
            False
            sage: from sage.combinat.posets.elements import PosetElement
            sage: PosetElement(P,0,"c") == PosetElement(P,0,"c")
            True
            sage: PosetElement(P,0,"c") == PosetElement(Q,0,"c")
            False
            sage: PosetElement(P,0,"b") == PosetElement(P,0,"c")
            False

        .. warning:: as an optimization, this only compares the parent
           and vertex, using the invariant that, in a proper poset
           element, ``self.element == other.element`` if and only
           ``self.vertex == other.vertex``::

            sage: PosetElement(P,1,"c") == PosetElement(P,0,"c")
            True

        Test that :issue:`12351` is fixed::

            sage: P(0) == int(0)
            False
        """
        # This should instead exploit unique representation, using
        # self is other, or best inherit __eq__ from there. But there
        # are issues around pickling and rich comparison functions.
        return have_same_parent(self, other) \
            and self.vertex == other.vertex

    def __ne__(self, other):
        r"""
        TESTS::

            sage: P = Poset([[1,2],[4],[3],[4],[]])
            sage: P = Poset([["a","b"],["d"],["c"],["d"],[]])
            sage: P(0).__ne__(P(4))
            True
            sage: from sage.combinat.posets.elements import PosetElement
            sage: PosetElement(P,0,"c") != PosetElement(P,0,"c")
            False
            sage: PosetElement(P,0,"b") != PosetElement(P,0,"c")
            True

        For this one, see comment in :meth:`__eq__`::

            sage: PosetElement(P,1,"c") != PosetElement(P,0,"c")
            False
        """
        return not self == other

    def _cmp(self, other):
        """
        TESTS::

            sage: P = Poset([[1,2],[4],[3],[4],[]], facade = False)
            sage: P(0)._cmp(P(4))
            -1
            sage: P(4)._cmp(P(0))
            1
            sage: P(0)._cmp(P(0))
            0
            sage: P(1)._cmp(P(2))
        """
        return self.parent().compare_elements(self, other)

    def __lt__(self, other):
        """
        TESTS::

            sage: dag = DiGraph({0:[2,3], 1:[3,4], 2:[5], 3:[5], 4:[5]})
            sage: P = Poset(dag, facade = False)
            sage: P(0) < P(1)
            False
            sage: P(4) < P(1)
            False
            sage: P(0) < P(0)
            False
        """
        return self._cmp(other) == -1 or False

    def __le__(self, other):
        """
        TESTS::

            sage: dag = DiGraph({0:[2,3], 1:[3,4], 2:[5], 3:[5], 4:[5]})
            sage: P = Poset(dag, facade = False)
            sage: P(1) <= P(0)
            False
            sage: P(0) <= P(1)
            False
            sage: P(0) <= P(3)
            True
            sage: P(0) <= P(0)
            True
        """
        return self == other or self._cmp(other) == -1 or False

    def __gt__(self, other):
        """
        TESTS::

            sage: dag = DiGraph({0:[2,3], 1:[3,4], 2:[5], 3:[5], 4:[5]})
            sage: P = Poset(dag)
            sage: P(0).__gt__(P(5))
            False
            sage: P(5).__gt__(P(0))
            True
            sage: P(0).__gt__(P(0))
            False
        """
        return self._cmp(other) == 1 or False

    def __ge__(self, other):
        """
        TESTS::

            sage: dag = DiGraph({0:[2,3], 1:[3,4], 2:[5], 3:[5], 4:[5]})
            sage: P = Poset(dag)
            sage: P(0).__ge__(P(5))
            False
            sage: P(5).__ge__(P(0))
            True
            sage: P(0).__ge__(P(0))
            True
        """
        return self == other or self._cmp(other) == 1 or False


class MeetSemilatticeElement(PosetElement):
    def __mul__(self, other):
        r"""
        Return the meet of ``self`` and ``other`` in the lattice.

        EXAMPLES::

            sage: D = posets.DiamondPoset(5, facade=False)
            sage: D(1) * D(2)
            0
            sage: D(1) * D(1)
            1
            sage: D(1) * D(0)
            0
            sage: D(1) * D(4)
            1
        """
        return self.parent().meet(self, other)


class JoinSemilatticeElement(PosetElement):
    def __add__(self, other):
        r"""
        Return the join of ``self`` and ``other`` in the lattice.

        EXAMPLES::

            sage: D = posets.DiamondPoset(5,facade=False)
            sage: D(1) + D(2)
            4
            sage: D(1) + D(1)
            1
            sage: D(1) + D(4)
            4
            sage: D(1) + D(0)
            1
        """
        return self.parent().join(self, other)


class LatticePosetElement(MeetSemilatticeElement, JoinSemilatticeElement):
    pass
